# flake8: noqa
import pkg_resources

from .config import config
from .type_sys import type_system
from .type_sys.utils import list_logical_types, list_semantic_tags
from .utils import concat_columns, read_file
from .version import __version__

import woodwork.column_accessor
import woodwork.demo
import woodwork.table_accessor
from woodwork.accessor_utils import (
    get_invalid_schema_message,
    init_series,
    is_schema_valid
)

# Call functions registered by other libraries when woodwork is imported
for entry_point in pkg_resources.iter_entry_points('alteryx_open_src_initialize'):  # pragma: no cover
    try:
        method = entry_point.load()
        if callable(method):
            method('woodwork')
    except Exception:
        pass



class DuplicateTagsWarning(UserWarning):
    def get_warning_message(self, duplicate_tags, name):
        return f"Semantic tag(s) '{', '.join(duplicate_tags)}' already present on column '{name}'"


class StandardTagsChangedWarning(UserWarning):
    def get_warning_message(self, use_standard_tags, col_name=None):
        changed = 'added to' if use_standard_tags else 'removed from'
        name = ("\"" + col_name + "\"") if col_name is not None else 'your column'
        return f'Standard tags have been {changed} {name}'


class UpgradeSchemaWarning(UserWarning):
    def get_warning_message(self, saved_version_str, current_schema_version):
        return ('The schema version of the saved Woodwork table '
                '%s is greater than the latest supported %s. '
                'You may need to upgrade woodwork. Attempting to load Woodwork table ...'
                % (saved_version_str, current_schema_version))


class OutdatedSchemaWarning(UserWarning):
    def get_warning_message(self, saved_version_str):
        return ('The schema version of the saved Woodwork table '
                '%s is no longer supported by this version '
                'of woodwork. Attempting to load Woodwork table ...'
                % (saved_version_str))


class IndexTagRemovedWarning(UserWarning):
    pass


class TypingInfoMismatchWarning(UserWarning):
    def get_warning_message(self, attr, invalid_reason, object_type):
        return (f'Operation performed by {attr} has invalidated the Woodwork typing information:\n '
                f'{invalid_reason}.\n '
                f'Please initialize Woodwork with {object_type}.ww.init')


class TypeConversionError(Exception):
    def __init__(self, series, new_dtype, logical_type):
        message = f'Error converting datatype for {series.name} from type {str(series.dtype)} '
        message += f'to type {new_dtype}. Please confirm the underlying data is consistent with '
        message += f'logical type {logical_type}.'
        super().__init__(message)


class ParametersIgnoredWarning(UserWarning):
    pass


class ColumnNotPresentError(KeyError):
    def __init__(self, column):
        if isinstance(column, str):
            return super().__init__(f"Column with name '{column}' not found in DataFrame")
        elif isinstance(column, list):
            return super().__init__(f"Column(s) '{column}' not found in DataFrame")


class WoodworkNotInitError(AttributeError):
    pass


class WoodworkNotInitWarning(UserWarning):
    pass



import re
from inspect import isclass

import numpy as np
import pandas as pd
import pytest
from mock import patch

import woodwork as ww
from woodwork.accessor_utils import (
    _is_dask_dataframe,
    _is_dask_series,
    _is_koalas_dataframe,
    _is_koalas_series,
    init_series
)
from woodwork.exceptions import (
    ColumnNotPresentError,
    IndexTagRemovedWarning,
    ParametersIgnoredWarning,
    TypeConversionError,
    TypingInfoMismatchWarning,
    WoodworkNotInitError
)
from woodwork.logical_types import (
    URL,
    Address,
    Age,
    AgeNullable,
    Boolean,
    BooleanNullable,
    Categorical,
    CountryCode,
    Datetime,
    Double,
    EmailAddress,
    Filepath,
    Integer,
    IntegerNullable,
    IPAddress,
    LatLong,
    NaturalLanguage,
    Ordinal,
    PersonFullName,
    PhoneNumber,
    PostalCode,
    SubRegionCode,
    Unknown
)
from woodwork.table_accessor import (
    WoodworkTableAccessor,
    _check_index,
    _check_logical_types,
    _check_time_index,
    _check_unique_column_names,
    _check_use_standard_tags
)
from woodwork.table_schema import TableSchema
from woodwork.tests.testing_utils import (
    is_property,
    is_public_method,
    to_pandas,
    validate_subset_schema
)
from woodwork.utils import import_or_none

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


def test_check_index_errors(sample_df):
    error_message = 'Specified index column `foo` not found in dataframe'
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_index(dataframe=sample_df, index='foo')

    if isinstance(sample_df, pd.DataFrame):
        # Does not check for index uniqueness with Dask
        error_message = 'Index column must be unique'
        with pytest.raises(LookupError, match=error_message):
            _check_index(sample_df, index='age')


def test_check_logical_types_errors(sample_df):
    error_message = 'logical_types must be a dictionary'
    with pytest.raises(TypeError, match=error_message):
        _check_logical_types(sample_df, logical_types='type')

    bad_logical_types_keys = {
        'full_name': None,
        'age': None,
        'birthday': None,
        'occupation': None,
    }
    error_message = re.escape("logical_types contains columns that are not present in dataframe: ['birthday', 'occupation']")
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_logical_types(sample_df, bad_logical_types_keys)


def test_check_time_index_errors(sample_df):
    error_message = 'Specified time index column `foo` not found in dataframe'
    with pytest.raises(ColumnNotPresentError, match=error_message):
        _check_time_index(dataframe=sample_df, time_index='foo')


def test_check_unique_column_names_errors(sample_df):
    if _is_koalas_dataframe(sample_df):
        pytest.skip("Koalas enforces unique column names")
    duplicate_cols_df = sample_df.copy()
    if _is_dask_dataframe(sample_df):
        duplicate_cols_df = dd.concat([duplicate_cols_df, duplicate_cols_df['age']], axis=1)
    else:
        duplicate_cols_df.insert(0, 'age', [18, 21, 65, 43], allow_duplicates=True)
    with pytest.raises(IndexError, match='Dataframe cannot contain duplicate columns names'):
        _check_unique_column_names(duplicate_cols_df)


def test_check_use_standard_tags_errors():
    error_message = 'use_standard_tags must be a dictionary or a boolean'
    with pytest.raises(TypeError, match=error_message):
        _check_use_standard_tags(1)


def test_accessor_init(sample_df):
    assert sample_df.ww.schema is None
    sample_df.ww.init()
    assert isinstance(sample_df.ww.schema, TableSchema)


def test_accessor_schema_property(sample_df):
    sample_df.ww.init()

    assert sample_df.ww._schema is not sample_df.ww.schema
    assert sample_df.ww._schema == sample_df.ww.schema


def test_set_accessor_name(sample_df):
    df = sample_df.copy()
    error = re.escape("Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init")
    with pytest.raises(WoodworkNotInitError, match=error):
        df.ww.name
    with pytest.raises(WoodworkNotInitError, match=error):
        df.ww.name = 'name'

    df.ww.init()
    assert df.ww.name is None
    df.ww.name = 'name'
    assert df.ww.schema.name == 'name'
    assert df.ww.name == 'name'


def test_rename_init_with_name(sample_df):
    df = sample_df.copy()
    df.ww.init(name='name')
    assert df.ww.name == 'name'
    df.ww.name = 'new_name'
    assert df.ww.schema.name == 'new_name'
    assert df.ww.name == 'new_name'


def test_name_error_on_init(sample_df):
    err_msg = "Table name must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_df.ww.init(name=123)


def test_name_error_on_update(sample_df):
    sample_df.ww.init()
    err_msg = "Table name must be a string"
    with pytest.raises(TypeError, match=err_msg):
        sample_df.ww.name = 123


def test_name_persists_after_drop(sample_df):
    df = sample_df.copy()
    df.ww.init()

    df.ww.name = 'name'
    assert df.ww.name == 'name'

    dropped_df = df.ww.drop(['id'])
    assert dropped_df.ww.name == 'name'
    assert dropped_df.ww.schema.name == 'name'


def test_set_accessor_metadata(sample_df):
    df = sample_df.copy()
    error = re.escape("Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init")
    with pytest.raises(WoodworkNotInitError, match=error):
        df.ww.metadata
    with pytest.raises(WoodworkNotInitError, match=error):
        df.ww.metadata = {'new': 'metadata'}

    df.ww.init()
    assert df.ww.metadata == {}
    df.ww.metadata = {'new': 'metadata'}
    assert df.ww.schema.metadata == {'new': 'metadata'}
    assert df.ww.metadata == {'new': 'metadata'}


def test_set_metadata_after_init_with_metadata(sample_df):
    df = sample_df.copy()
    df.ww.init(table_metadata={'new': 'metadata'})
    assert df.ww.metadata == {'new': 'metadata'}
    df.ww.metadata = {'new': 'new_metadata'}
    assert df.ww.schema.metadata == {'new': 'new_metadata'}
    assert df.ww.metadata == {'new': 'new_metadata'}


def test_metadata_persists_after_drop(sample_df):
    df = sample_df.copy()
    df.ww.init()

    df.ww.metadata = {'new': 'metadata'}
    assert df.ww.metadata == {'new': 'metadata'}

    dropped_df = df.ww.drop(['id'])
    assert dropped_df.ww.metadata == {'new': 'metadata'}
    assert dropped_df.ww.schema.metadata == {'new': 'metadata'}


def test_metadata_error_on_init(sample_df):
    err_msg = 'Table metadata must be a dictionary.'
    with pytest.raises(TypeError, match=err_msg):
        sample_df.ww.init(table_metadata=123)


def test_metadata_error_on_update(sample_df):
    sample_df.ww.init()
    err_msg = 'Table metadata must be a dictionary.'
    with pytest.raises(TypeError, match=err_msg):
        sample_df.ww.metadata = 123


def test_accessor_physical_types_property(sample_df):
    sample_df.ww.init(logical_types={'age': 'Categorical'})

    assert isinstance(sample_df.ww.physical_types, dict)
    assert set(sample_df.ww.physical_types.keys()) == set(sample_df.columns)
    for k, v in sample_df.ww.physical_types.items():
        logical_type = sample_df.ww.columns[k].logical_type
        if _is_koalas_dataframe(sample_df) and logical_type.backup_dtype is not None:
            assert v == logical_type.backup_dtype
        else:
            assert v == logical_type.primary_dtype


def test_accessor_separation_of_params(sample_df):
    # mix up order of acccessor and schema params
    schema_df = sample_df.copy()
    schema_df.ww.init(name='test_name', index='id', semantic_tags={'id': 'test_tag'}, time_index='signup_date')

    assert schema_df.ww.semantic_tags['id'] == {'index', 'test_tag'}
    assert schema_df.ww.index == 'id'
    assert schema_df.ww.time_index == 'signup_date'
    assert schema_df.ww.name == 'test_name'


def test_init_accessor_with_schema(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(name='test_schema', semantic_tags={'id': 'test_tag'}, index='id')
    schema = schema_df.ww._schema

    head_df = schema_df.head(2)
    assert head_df.ww.schema is None
    head_df.ww.init(schema=schema)

    assert head_df.ww._schema is schema
    assert head_df.ww.name == 'test_schema'
    assert head_df.ww.semantic_tags['id'] == {'index', 'test_tag'}

    iloc_df = schema_df.loc[[2, 3]]
    assert iloc_df.ww.schema is None
    iloc_df.ww.init(schema=schema)

    assert iloc_df.ww._schema is schema
    assert iloc_df.ww.name == 'test_schema'
    assert iloc_df.ww.semantic_tags['id'] == {'index', 'test_tag'}
    # Extra parameters do not take effect
    assert isinstance(iloc_df.ww.logical_types['id'], Integer)


def test_accessor_init_errors_methods(sample_df):
    methods_to_exclude = ['init']
    public_methods = [method for method in dir(sample_df.ww) if is_public_method(WoodworkTableAccessor, method)]
    public_methods = [method for method in public_methods if method not in methods_to_exclude]
    method_args_dict = {
        'add_semantic_tags': [{'id': 'new_tag'}],
        'describe': None,
        'pop': ['id'],
        'describe': None,
        'describe_dict': None,
        'drop': ['id'],
        'mutual_information': None,
        'mutual_information_dict': None,
        'remove_semantic_tags': [{'id': 'new_tag'}],
        'rename': [{'id': 'new_id'}],
        'reset_semantic_tags': None,
        'select': [['Double']],
        'set_index': ['id'],
        'set_time_index': ['signup_date'],
        'set_types': [{'id': 'Integer'}],
        'to_disk': ['dir'],
        'to_dictionary': None,
        'value_counts': None,

    }
    error = re.escape("Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init")
    for method in public_methods:
        func = getattr(sample_df.ww, method)
        method_args = method_args_dict[method]
        with pytest.raises(WoodworkNotInitError, match=error):
            if method_args:
                func(*method_args)
            else:
                func()


def test_accessor_init_errors_properties(sample_df):
    props_to_exclude = ['iloc', 'loc', 'schema', '_dataframe']
    props = [prop for prop in dir(sample_df.ww) if is_property(WoodworkTableAccessor, prop) and prop not in props_to_exclude]

    error = re.escape("Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init")
    for prop in props:
        with pytest.raises(WoodworkNotInitError, match=error):
            getattr(sample_df.ww, prop)


def test_init_accessor_with_schema_errors(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init()
    schema = schema_df.ww.schema

    iloc_df = schema_df.iloc[:, :-1]
    assert iloc_df.ww.schema is None

    error = 'Provided schema must be a Woodwork.TableSchema object.'
    with pytest.raises(TypeError, match=error):
        iloc_df.ww.init(schema=int)

    error = ("Woodwork typing information is not valid for this DataFrame: "
             "The following columns in the typing information were missing from the DataFrame: {'datetime_with_NaT'}")
    with pytest.raises(ValueError, match=error):
        iloc_df.ww.init(schema=schema)


def test_accessor_with_schema_parameter_warning(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(name='test_schema', semantic_tags={'id': 'test_tag'}, index='id')
    schema = schema_df.ww.schema

    head_df = schema_df.head(2)

    warning = "A schema was provided and the following parameters were ignored: index, " \
              "time_index, logical_types, already_sorted, use_standard_tags, semantic_tags"
    with pytest.warns(ParametersIgnoredWarning, match=warning):
        head_df.ww.init(index='ignored_id', time_index="ignored_time_index", logical_types={'ignored': 'ltypes'},
                        already_sorted=True, semantic_tags={'ignored_id': 'ignored_test_tag'},
                        use_standard_tags={'id': True, 'age': False}, schema=schema)

    assert head_df.ww.name == 'test_schema'
    assert head_df.ww.semantic_tags['id'] == {'index', 'test_tag'}


def test_accessor_getattr(sample_df):
    schema_df = sample_df.copy()

    # We can access attributes on the Accessor class before the schema is initialized
    assert schema_df.ww.schema is None

    error = re.escape("Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init")
    with pytest.raises(WoodworkNotInitError, match=error):
        schema_df.ww.index

    schema_df.ww.init()

    assert schema_df.ww.name is None
    assert schema_df.ww.index is None
    assert schema_df.ww.time_index is None

    assert set(schema_df.ww.columns.keys()) == set(sample_df.columns)

    error = re.escape("Woodwork has no attribute 'not_present'")
    with pytest.raises(AttributeError, match=error):
        sample_df.ww.init()
        sample_df.ww.not_present


def test_getitem(sample_df):
    df = sample_df
    df.ww.init(
        time_index='signup_date',
        index='id', name='df_name',
        logical_types={'age': 'Double'},
        semantic_tags={'age': {'custom_tag'}},
    )
    assert list(df.columns) == list(df.ww.schema.columns)

    subset = ['id', 'signup_date']
    df_subset = df.ww[subset]
    pd.testing.assert_frame_equal(to_pandas(df[subset]), to_pandas(df_subset))
    assert subset == list(df_subset.ww._schema.columns)
    assert df_subset.ww.index == 'id'
    assert df_subset.ww.time_index == 'signup_date'

    subset = ['age', 'email']
    df_subset = df.ww[subset]
    pd.testing.assert_frame_equal(to_pandas(df[subset]), to_pandas(df_subset))
    assert subset == list(df_subset.ww._schema.columns)
    assert df_subset.ww.index is None
    assert df_subset.ww.time_index is None
    assert isinstance(df_subset.ww.logical_types['age'], Double)
    assert df_subset.ww.semantic_tags['age'] == {'custom_tag', 'numeric'}

    subset = df.ww[[]]
    assert len(subset.ww.columns) == 0
    assert subset.ww.index is None
    assert subset.ww.time_index is None

    series = df.ww['age']
    pd.testing.assert_series_equal(to_pandas(series), to_pandas(df['age']))
    assert isinstance(series.ww.logical_type, Double)
    assert series.ww.semantic_tags == {'custom_tag', 'numeric'}

    series = df.ww['id']
    pd.testing.assert_series_equal(to_pandas(series), to_pandas(df['id']))
    assert isinstance(series.ww.logical_type, Integer)
    assert series.ww.semantic_tags == {'index'}


def test_getitem_init_error(sample_df):
    error = re.escape("Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init")
    with pytest.raises(WoodworkNotInitError, match=error):
        sample_df.ww['age']


def test_getitem_invalid_input(sample_df):
    df = sample_df
    df.ww.init()

    error_msg = r"Column\(s\) '\[1, 2\]' not found in DataFrame"
    with pytest.raises(ColumnNotPresentError, match=error_msg):
        df.ww[['email', 2, 1]]

    error_msg = "Column with name 'invalid_column' not found in DataFrame"
    with pytest.raises(ColumnNotPresentError, match=error_msg):
        df.ww['invalid_column']


def test_accessor_equality(sample_df):
    # Confirm equality with same schema and same data
    schema_df = sample_df.copy()
    schema_df.ww.init()

    copy_df = schema_df.ww.copy()
    assert schema_df.ww == copy_df.ww

    # Confirm not equal with different schema but same data
    copy_df.ww.set_time_index('signup_date')
    assert schema_df.ww != copy_df.ww

    # Confirm not equal with same schema but different data - only pandas
    loc_df = schema_df.ww.loc[:2, :]
    if isinstance(sample_df, pd.DataFrame):
        assert schema_df.ww != loc_df
    else:
        assert schema_df.ww == loc_df


def test_accessor_shallow_equality(sample_df):
    metadata_table = sample_df.copy()
    metadata_table.ww.init(table_metadata={'user': 'user0'})
    diff_metadata_table = sample_df.copy()
    diff_metadata_table.ww.init(table_metadata={'user': 'user2'})

    assert diff_metadata_table.ww.__eq__(metadata_table, deep=False)
    assert not diff_metadata_table.ww.__eq__(metadata_table, deep=True)

    schema = metadata_table.ww.schema
    diff_data_table = metadata_table.ww.loc[:2, :]
    same_data_table = metadata_table.ww.copy()

    assert diff_data_table.ww.schema.__eq__(schema, deep=True)
    assert same_data_table.ww.schema.__eq__(schema, deep=True)

    assert same_data_table.ww.__eq__(metadata_table.ww, deep=False)
    assert same_data_table.ww.__eq__(metadata_table.ww, deep=True)

    assert diff_data_table.ww.__eq__(metadata_table.ww, deep=False)
    if isinstance(sample_df, pd.DataFrame):
        assert not diff_data_table.ww.__eq__(metadata_table.ww, deep=True)


def test_accessor_init_with_valid_string_time_index(time_index_df):
    time_index_df.ww.init(name='schema',
                          index='id',
                          time_index='times')

    assert time_index_df.ww.name == 'schema'
    assert time_index_df.ww.index == 'id'
    assert time_index_df.ww.time_index == 'times'
    assert isinstance(time_index_df.ww.columns[time_index_df.ww.time_index].logical_type, Datetime)


def test_accessor_init_with_numeric_datetime_time_index(time_index_df):
    schema_df = time_index_df.copy()
    schema_df.ww.init(time_index='ints', logical_types={'ints': Datetime})

    error_msg = 'Time index column must contain datetime or numeric values'
    with pytest.raises(TypeError, match=error_msg):
        time_index_df.ww.init(name='schema', time_index='strs', logical_types={'strs': Datetime})

    assert schema_df.ww.time_index == 'ints'
    assert schema_df['ints'].dtype == 'datetime64[ns]'


def test_accessor_with_numeric_time_index(time_index_df):
    # Set a numeric time index on init
    schema_df = time_index_df.copy()
    schema_df.ww.init(time_index='ints')
    date_col = schema_df.ww.columns['ints']
    assert schema_df.ww.time_index == 'ints'
    assert isinstance(date_col.logical_type, Integer)
    assert date_col.semantic_tags == {'time_index', 'numeric'}

    # Specify logical type for time index on init
    schema_df = time_index_df.copy()
    schema_df.ww.init(time_index='ints', logical_types={'ints': 'Double'})
    date_col = schema_df.ww.columns['ints']
    assert schema_df.ww.time_index == 'ints'
    assert isinstance(date_col.logical_type, Double)
    assert date_col.semantic_tags == {'time_index', 'numeric'}

    schema_df = time_index_df.copy()
    schema_df.ww.init(time_index='strs', logical_types={'strs': 'Double'})
    date_col = schema_df.ww.columns['strs']
    assert schema_df.ww.time_index == 'strs'
    assert isinstance(date_col.logical_type, Double)
    assert date_col.semantic_tags == {'time_index', 'numeric'}

    error_msg = 'Time index column must contain datetime or numeric values'
    with pytest.raises(TypeError, match=error_msg):
        time_index_df.ww.init(time_index='ints', logical_types={'ints': 'Categorical'})

    error_msg = 'Time index column must contain datetime or numeric values'
    with pytest.raises(TypeError, match=error_msg):
        time_index_df.ww.init(time_index='letters', logical_types={'strs': 'Integer'})

    # Set numeric time index after init
    schema_df = time_index_df.copy()
    schema_df.ww.init(logical_types={'ints': 'Double'})
    assert schema_df.ww.time_index is None

    schema_df.ww.set_time_index('ints')
    date_col = schema_df.ww.columns['ints']
    assert schema_df.ww.time_index == 'ints'
    assert isinstance(date_col.logical_type, Double)
    assert date_col.semantic_tags == {'numeric', 'time_index'}


def test_numeric_time_index_dtypes(numeric_time_index_df):
    numeric_time_index_df.ww.init(time_index='ints')
    assert numeric_time_index_df.ww.time_index == 'ints'
    assert isinstance(numeric_time_index_df.ww.logical_types['ints'], Integer)
    assert numeric_time_index_df.ww.semantic_tags['ints'] == {'time_index', 'numeric'}

    numeric_time_index_df.ww.set_time_index('floats')
    assert numeric_time_index_df.ww.time_index == 'floats'
    assert isinstance(numeric_time_index_df.ww.logical_types['floats'], Double)
    assert numeric_time_index_df.ww.semantic_tags['floats'] == {'time_index', 'numeric'}

    numeric_time_index_df.ww.set_time_index('with_null')
    assert numeric_time_index_df.ww.time_index == 'with_null'
    assert isinstance(numeric_time_index_df.ww.logical_types['with_null'], IntegerNullable)
    assert numeric_time_index_df.ww.semantic_tags['with_null'] == {'time_index', 'numeric'}


def test_accessor_init_with_invalid_string_time_index(sample_df):
    error_msg = 'Time index column must contain datetime or numeric values'
    with pytest.raises(TypeError, match=error_msg):
        sample_df.ww.init(name='schema', time_index='full_name')


def test_accessor_init_with_string_logical_types(sample_df):
    logical_types = {
        'full_name': 'natural_language',
        'age': 'Double'
    }
    schema_df = sample_df.copy()
    schema_df.ww.init(name='schema',
                      logical_types=logical_types)
    assert isinstance(schema_df.ww.columns['full_name'].logical_type, NaturalLanguage)
    assert isinstance(schema_df.ww.columns['age'].logical_type, Double)

    logical_types = {
        'full_name': 'NaturalLanguage',
        'age': 'IntegerNullable',
        'signup_date': 'Datetime'
    }
    schema_df = sample_df.copy()
    schema_df.ww.init(name='schema',
                      logical_types=logical_types,
                      time_index='signup_date'
                      )
    assert isinstance(schema_df.ww.columns['full_name'].logical_type, NaturalLanguage)
    assert isinstance(schema_df.ww.columns['age'].logical_type, IntegerNullable)
    assert schema_df.ww.time_index == 'signup_date'


def test_int_dtype_inference_on_init():
    df = pd.DataFrame({
        'ints_no_nans': pd.Series([1, 2]),
        'ints_nan': pd.Series([1, np.nan]),
        'ints_NA': pd.Series([1, pd.NA]),
        'ints_NA_specified': pd.Series([1, pd.NA], dtype='Int64')})
    df = df.loc[df.index.repeat(5)].reset_index(drop=True)
    df.ww.init()

    assert df['ints_no_nans'].dtype == 'int64'
    assert df['ints_nan'].dtype == 'float64'
    assert df['ints_NA'].dtype == 'category'
    assert df['ints_NA_specified'].dtype == 'Int64'


def test_bool_dtype_inference_on_init():
    df = pd.DataFrame({
        'bools_no_nans': pd.Series([True, False]),
        'bool_nan': pd.Series([True, np.nan]),
        'bool_NA': pd.Series([True, pd.NA]),
        'bool_NA_specified': pd.Series([True, pd.NA], dtype="boolean")})
    df = df.loc[df.index.repeat(5)].reset_index(drop=True)
    df.ww.init()

    assert df['bools_no_nans'].dtype == 'bool'
    assert df['bool_nan'].dtype == 'category'
    assert df['bool_NA'].dtype == 'category'
    assert df['bool_NA_specified'].dtype == 'boolean'


def test_str_dtype_inference_on_init():
    df = pd.DataFrame({
        'str_no_nans': pd.Series(['a', 'b']),
        'str_nan': pd.Series(['a', np.nan]),
        'str_NA': pd.Series(['a', pd.NA]),
        'str_NA_specified': pd.Series([1, pd.NA], dtype="string"),
    })
    df = df.loc[df.index.repeat(5)].reset_index(drop=True)
    df.ww.init()

    assert df['str_no_nans'].dtype == 'category'
    assert df['str_nan'].dtype == 'category'
    assert df['str_NA'].dtype == 'category'
    assert df['str_NA_specified'].dtype == 'category'


def test_float_dtype_inference_on_init():
    df = pd.DataFrame({
        'floats_no_nans': pd.Series([1.1, 2.2]),
        'floats_nan': pd.Series([1.1, np.nan]),
        'floats_NA': pd.Series([1.1, pd.NA]),
        'floats_nan_specified': pd.Series([1.1, np.nan], dtype='float')})
    df = df.loc[df.index.repeat(5)].reset_index(drop=True)
    df.ww.init()

    assert df['floats_no_nans'].dtype == 'float64'
    assert df['floats_nan'].dtype == 'float64'
    assert df['floats_NA'].dtype == 'category'
    assert df['floats_nan_specified'].dtype == 'float64'


def test_datetime_dtype_inference_on_init():
    df = pd.DataFrame({
        'date_no_nans': pd.Series([pd.to_datetime('2020-09-01')] * 2),
        'date_nan': pd.Series([pd.to_datetime('2020-09-01'), np.nan]),
        'date_NA': pd.Series([pd.to_datetime('2020-09-01'), pd.NA]),
        'date_NaT': pd.Series([pd.to_datetime('2020-09-01'), pd.NaT]),
        'date_NA_specified': pd.Series([pd.to_datetime('2020-09-01'), pd.NA], dtype='datetime64[ns]')})
    df.ww.init()

    assert df['date_no_nans'].dtype == 'datetime64[ns]'
    assert df['date_nan'].dtype == 'datetime64[ns]'
    assert df['date_NA'].dtype == 'datetime64[ns]'
    assert df['date_NaT'].dtype == 'datetime64[ns]'
    assert df['date_NA_specified'].dtype == 'datetime64[ns]'


def test_datetime_inference_with_format_param():
    df = pd.DataFrame({
        'index': [0, 1, 2],
        'dates': ["2019/01/01", "2019/01/02", "2019/01/03"],
        'ymd_special': ["2019~01~01", "2019~01~02", "2019~01~03"],
        'mdy_special': pd.Series(['3~11~2000', '3~12~2000', '3~13~2000'], dtype='string'),
    })
    df.ww.init(
        name='df_name',
        logical_types={'ymd_special': Datetime(datetime_format='%Y~%m~%d'),
                       'mdy_special': Datetime(datetime_format='%m~%d~%Y'),
                       'dates': Datetime},
        time_index='ymd_special')

    assert df['dates'].dtype == 'datetime64[ns]'
    assert df['ymd_special'].dtype == 'datetime64[ns]'
    assert df['mdy_special'].dtype == 'datetime64[ns]'

    assert df.ww.time_index == 'ymd_special'
    assert isinstance(df.ww['dates'].ww.logical_type, Datetime)
    assert isinstance(df.ww['ymd_special'].ww.logical_type, Datetime)
    assert isinstance(df.ww['mdy_special'].ww.logical_type, Datetime)

    df.ww.set_time_index('mdy_special')
    assert df.ww.time_index == 'mdy_special'

    df = pd.DataFrame({
        'mdy_special': pd.Series(['3&11&2000', '3&12&2000', '3&13&2000'], dtype='string'),
    })
    df = df.loc[df.index.repeat(5)].reset_index(drop=True)
    df.ww.init()
    assert df['mdy_special'].dtype == 'category'

    df.ww.set_types(logical_types={'mdy_special': Datetime(datetime_format='%m&%d&%Y')})
    assert df['mdy_special'].dtype == 'datetime64[ns]'

    df.ww.set_time_index('mdy_special')
    assert isinstance(df.ww['mdy_special'].ww.logical_type, Datetime)
    assert df.ww.time_index == 'mdy_special'


def test_timedelta_dtype_inference_on_init():
    df = pd.DataFrame({
        'delta_no_nans': (pd.Series([pd.to_datetime('2020-09-01')] * 2) - pd.to_datetime('2020-07-01')),
        'delta_nan': (pd.Series([pd.to_datetime('2020-09-01'), np.nan]) - pd.to_datetime('2020-07-01')),
        'delta_NaT': (pd.Series([pd.to_datetime('2020-09-01'), pd.NaT]) - pd.to_datetime('2020-07-01')),
        'delta_NA_specified': (pd.Series([pd.to_datetime('2020-09-01'), pd.NA], dtype='datetime64[ns]') - pd.to_datetime('2020-07-01')),
    })
    df.ww.init()

    assert df['delta_no_nans'].dtype == 'timedelta64[ns]'
    assert df['delta_nan'].dtype == 'timedelta64[ns]'
    assert df['delta_NaT'].dtype == 'timedelta64[ns]'
    assert df['delta_NA_specified'].dtype == 'timedelta64[ns]'


def test_sets_category_dtype_on_init():
    column_name = 'test_series'
    series_list = [
        pd.Series(['a', 'b', 'c'], name=column_name),
        pd.Series(['a', None, 'c'], name=column_name),
        pd.Series(['a', np.nan, 'c'], name=column_name),
        pd.Series(['a', pd.NA, 'c'], name=column_name),
        pd.Series(['a', pd.NaT, 'c'], name=column_name),
    ]

    logical_types = [
        Categorical,
        CountryCode,
        Ordinal(order=['a', 'b', 'c']),
        PostalCode,
        SubRegionCode
    ]

    for series in series_list:
        series = series.astype('object')
        for logical_type in logical_types:
            if isclass(logical_type):
                logical_type = logical_type()

            ltypes = {
                column_name: logical_type,
            }
            df = pd.DataFrame(series)
            df.ww.init(logical_types=ltypes)
            assert df.ww.columns[column_name].logical_type == logical_type
            assert df[column_name].dtype == logical_type.primary_dtype


def test_sets_object_dtype_on_init(latlong_df):
    for column_name in latlong_df.columns:
        ltypes = {
            column_name: LatLong,
        }
        df = latlong_df.loc[:, [column_name]]
        df.ww.init(logical_types=ltypes)
        assert isinstance(df.ww.columns[column_name].logical_type, LatLong)
        assert df[column_name].dtype == LatLong.primary_dtype
        df_pandas = to_pandas(df[column_name])
        expected_val = (3, 4)
        if _is_koalas_dataframe(latlong_df):
            expected_val = [3, 4]
        assert df_pandas.iloc[-1] == expected_val


def test_sets_string_dtype_on_init():
    column_name = 'test_series'
    series_list = [
        pd.Series(['a', 'b', 'c'], name=column_name),
        pd.Series(['a', None, 'c'], name=column_name),
        pd.Series(['a', np.nan, 'c'], name=column_name),
        pd.Series(['a', pd.NA, 'c'], name=column_name),
    ]

    logical_types = [
        Address,
        Filepath,
        PersonFullName,
        IPAddress,
        NaturalLanguage,
        PhoneNumber,
        URL,
    ]

    for series in series_list:
        series = series.astype('object')
        for logical_type in logical_types:
            ltypes = {
                column_name: logical_type,
            }
            df = pd.DataFrame(series)
            df.ww.init(logical_types=ltypes)
            assert isinstance(df.ww.columns[column_name].logical_type, logical_type)
            assert df[column_name].dtype == logical_type.primary_dtype


def test_sets_boolean_dtype_on_init():
    column_name = 'test_series'
    series_list = [
        pd.Series([True, False, True], name=column_name),
        pd.Series([True, None, True], name=column_name),
        pd.Series([True, np.nan, True], name=column_name),
        pd.Series([True, pd.NA, True], name=column_name),
    ]

    logical_types = [Boolean, BooleanNullable]
    for series in series_list:
        for logical_type in logical_types:
            if series.isnull().any() and logical_type == Boolean:
                continue
            series = series.astype('object')
            ltypes = {
                column_name: logical_type,
            }
            df = pd.DataFrame(series)
            df.ww.init(logical_types=ltypes)
            assert isinstance(df.ww.columns[column_name].logical_type, logical_type)
            assert df[column_name].dtype == logical_type.primary_dtype


def test_sets_int64_dtype_on_init():
    column_name = 'test_series'
    series_list = [
        pd.Series([1, 2, 3], name=column_name),
        pd.Series([1, None, 3], name=column_name),
        pd.Series([1, np.nan, 3], name=column_name),
        pd.Series([1, pd.NA, 3], name=column_name),
    ]

    logical_types = [Integer, IntegerNullable, Age, AgeNullable]
    for series in series_list:
        series = series.astype('object')
        for logical_type in logical_types:
            if series.isnull().any() and logical_type in [Integer, Age]:
                continue
            ltypes = {
                column_name: logical_type,
            }
            df = pd.DataFrame(series)
            df.ww.init(logical_types=ltypes)
            assert isinstance(df.ww.columns[column_name].logical_type, logical_type)
            assert df[column_name].dtype == logical_type.primary_dtype


def test_sets_float64_dtype_on_init():
    column_name = 'test_series'
    series_list = [
        pd.Series([1.1, 2, 3], name=column_name),
        pd.Series([1.1, None, 3], name=column_name),
        pd.Series([1.1, np.nan, 3], name=column_name),
    ]

    logical_type = Double
    for series in series_list:
        series = series.astype('object')
        ltypes = {
            column_name: logical_type,
        }
        df = pd.DataFrame(series)
        df.ww.init(logical_types=ltypes)
        assert isinstance(df.ww.columns[column_name].logical_type, logical_type)
        assert df[column_name].dtype == logical_type.primary_dtype


def test_sets_datetime64_dtype_on_init():
    column_name = 'test_series'
    series_list = [
        pd.Series(['2020-01-01', '2020-01-02', '2020-01-03'], name=column_name),
        pd.Series(['2020-01-01', None, '2020-01-03'], name=column_name),
        pd.Series(['2020-01-01', np.nan, '2020-01-03'], name=column_name),
        pd.Series(['2020-01-01', pd.NA, '2020-01-03'], name=column_name),
        pd.Series(['2020-01-01', pd.NaT, '2020-01-03'], name=column_name, dtype='object'),
    ]

    logical_type = Datetime
    for series in series_list:
        series = series.astype('object')
        ltypes = {
            column_name: logical_type,
        }
        df = pd.DataFrame(series)
        df.ww.init(logical_types=ltypes)
        assert isinstance(df.ww.columns[column_name].logical_type, logical_type)
        assert df[column_name].dtype == logical_type.primary_dtype


def test_invalid_dtype_casting():
    column_name = 'test_series'

    # Cannot cast a column with pd.NA to Double
    series = pd.Series([1.1, pd.NA, 3], name=column_name)
    ltypes = {
        column_name: Double,
    }

    err_msg = 'Error converting datatype for test_series from type object to type ' \
        'float64. Please confirm the underlying data is consistent with logical type Double.'
    df = pd.DataFrame(series)
    with pytest.raises(TypeConversionError, match=err_msg):
        df.ww.init(logical_types=ltypes)

    # Cannot cast Datetime to Double
    df = pd.DataFrame({column_name: ['2020-01-01', '2020-01-02', '2020-01-03']})
    df.ww.init(logical_types={column_name: Datetime})

    err_msg = 'Error converting datatype for test_series from type datetime64[ns] to type ' \
        'float64. Please confirm the underlying data is consistent with logical type Double.'
    with pytest.raises(TypeConversionError, match=re.escape(err_msg)):
        df.ww.set_types(logical_types={column_name: Double})

    # Cannot cast invalid strings to integers
    series = pd.Series(['1', 'two', '3'], name=column_name)
    ltypes = {
        column_name: Integer,
    }
    err_msg = 'Error converting datatype for test_series from type object to type ' \
        'int64. Please confirm the underlying data is consistent with logical type Integer.'
    df = pd.DataFrame(series)
    with pytest.raises(TypeConversionError, match=err_msg):
        df.ww.init(logical_types=ltypes)


def test_underlying_index_set_no_index_on_init(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('Setting underlying index is not supported with Dask input')
    if _is_koalas_dataframe(sample_df):
        pytest.xfail('Setting underlying index is not supported with Koalas input')

    input_index = pd.Int64Index([99, 88, 77, 66])

    schema_df = sample_df.copy()
    schema_df.index = input_index.copy()
    pd.testing.assert_index_equal(input_index, schema_df.index)

    schema_df.ww.init()
    assert schema_df.ww.index is None
    pd.testing.assert_index_equal(input_index, schema_df.index)

    sorted_df = schema_df.ww.sort_values('full_name')
    assert sorted_df.ww.index is None
    pd.testing.assert_index_equal(pd.Int64Index([88, 77, 99, 66]), sorted_df.index)


def test_underlying_index_set(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('Setting underlying index is not supported with Dask input')
    if _is_koalas_dataframe(sample_df):
        pytest.xfail('Setting underlying index is not supported with Koalas input')

    # Sets underlying index at init
    schema_df = sample_df.copy()
    schema_df.ww.init(index='full_name')
    assert 'full_name' in schema_df.columns
    assert schema_df.index.name is None
    assert (schema_df.index == schema_df['full_name']).all()

    # Sets underlying index on update
    schema_df = sample_df.copy()
    schema_df.ww.init(index='id')
    schema_df.ww.set_index('full_name')
    assert schema_df.ww.index == 'full_name'
    assert 'full_name' in schema_df.columns
    assert (schema_df.index == schema_df['full_name']).all()
    assert schema_df.index.name is None

    # confirm removing Woodwork index doesn't change underlying index
    schema_df.ww.set_index(None)
    assert schema_df.ww.index is None
    assert (schema_df.index == schema_df['full_name']).all()


def test_underlying_index_reset(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('Setting underlying index is not supported with Dask input')
    if _is_koalas_dataframe(sample_df):
        pytest.xfail('Setting underlying index is not supported with Koalas input')

    specified_index = pd.Index
    unspecified_index = pd.RangeIndex

    sample_df.ww.init()
    assert type(sample_df.index) == unspecified_index

    sample_df.ww.set_index('full_name')
    assert type(sample_df.index) == specified_index

    copied_df = sample_df.ww.copy()
    warning = ("Index mismatch between DataFrame and typing information")
    with pytest.warns(TypingInfoMismatchWarning, match=warning):
        copied_df.ww.reset_index(drop=True, inplace=True)
    assert copied_df.ww.schema is None
    assert type(copied_df.index) == unspecified_index

    sample_df.ww.set_index(None)
    assert type(sample_df.index) == specified_index

    # Use pandas operation to reset index
    reset_df = sample_df.ww.reset_index(drop=True, inplace=False)
    assert type(sample_df.index) == specified_index
    assert type(reset_df.index) == unspecified_index

    sample_df.ww.reset_index(drop=True, inplace=True)
    assert type(sample_df.index) == unspecified_index


def test_underlying_index_unchanged_after_updates(sample_df):
    if _is_dask_dataframe(sample_df):
        pytest.xfail('Setting underlying index is not supported with Dask input')
    if _is_koalas_dataframe(sample_df):
        pytest.xfail('Setting underlying index is not supported with Koalas input')

    sample_df.ww.init(index='full_name')
    assert 'full_name' in sample_df
    assert sample_df.ww.index == 'full_name'
    assert (sample_df.index == sample_df['full_name']).all()

    copied_df = sample_df.ww.copy()

    dropped_df = copied_df.ww.drop('full_name')
    assert 'full_name' not in dropped_df
    assert dropped_df.ww.index is None
    assert (dropped_df.index == sample_df['full_name']).all()

    selected_df = copied_df.ww.select('Integer')
    assert 'full_name' not in dropped_df
    assert selected_df.ww.index is None
    assert (selected_df.index == sample_df['full_name']).all()

    iloc_df = copied_df.ww.iloc[:, 2:]
    assert 'full_name' not in iloc_df
    assert iloc_df.ww.index is None
    assert (iloc_df.index == sample_df['full_name']).all()

    loc_df = copied_df.ww.loc[:, ['id', 'email']]
    assert 'full_name' not in loc_df
    assert loc_df.ww.index is None
    assert (loc_df.index == sample_df['full_name']).all()

    subset_df = copied_df.ww[['id', 'email']]
    assert 'full_name' not in subset_df
    assert subset_df.ww.index is None
    assert (subset_df.index == sample_df['full_name']).all()

    reset_tags_df = sample_df.ww.copy()
    reset_tags_df.ww.reset_semantic_tags('full_name', retain_index_tags=False)
    assert reset_tags_df.ww.index is None
    assert (reset_tags_df.index == sample_df['full_name']).all()

    remove_tags_df = sample_df.ww.copy()
    remove_tags_df.ww.remove_semantic_tags({'full_name': 'index'})
    assert remove_tags_df.ww.index is None
    assert (remove_tags_df.index == sample_df['full_name']).all()

    set_types_df = sample_df.ww.copy()
    set_types_df.ww.set_types(semantic_tags={'full_name': 'new_tag'}, retain_index_tags=False)
    assert set_types_df.ww.index is None
    assert (set_types_df.index == sample_df['full_name']).all()

    popped_df = sample_df.ww.copy()
    popped_df.ww.pop('full_name')
    assert popped_df.ww.index is None
    assert (popped_df.index == sample_df['full_name']).all()


def test_accessor_already_sorted(sample_unsorted_df):
    if _is_dask_dataframe(sample_unsorted_df):
        pytest.xfail('Sorting dataframe is not supported with Dask input')
    if _is_koalas_dataframe(sample_unsorted_df):
        pytest.xfail('Sorting dataframe is not supported with Koalas input')

    schema_df = sample_unsorted_df.copy()
    schema_df.ww.init(name='schema',
                      index='id',
                      time_index='signup_date')

    assert schema_df.ww.time_index == 'signup_date'
    assert isinstance(schema_df.ww.columns[schema_df.ww.time_index].logical_type, Datetime)

    sorted_df = to_pandas(sample_unsorted_df).sort_values(['signup_date', 'id']).set_index('id', drop=False)
    sorted_df.index.name = None
    pd.testing.assert_frame_equal(sorted_df,
                                  to_pandas(schema_df), check_index_type=False, check_dtype=False)

    schema_df = sample_unsorted_df.copy()
    schema_df.ww.init(name='schema',
                      index='id',
                      time_index='signup_date',
                      already_sorted=True)

    assert schema_df.ww.time_index == 'signup_date'
    assert isinstance(schema_df.ww.columns[schema_df.ww.time_index].logical_type, Datetime)

    unsorted_df = to_pandas(sample_unsorted_df.set_index('id', drop=False))
    unsorted_df.index.name = None
    pd.testing.assert_frame_equal(unsorted_df, to_pandas(schema_df), check_index_type=False, check_dtype=False)


def test_ordinal_with_order(sample_series):
    if _is_koalas_series(sample_series) or _is_dask_series(sample_series):
        pytest.xfail('Fails with Dask and Koalas - ordinal data validation not compatible')

    ordinal_with_order = Ordinal(order=['a', 'b', 'c'])
    schema_df = pd.DataFrame(sample_series)
    schema_df.ww.init(logical_types={'sample_series': ordinal_with_order})

    column_logical_type = schema_df.ww.logical_types['sample_series']
    assert isinstance(column_logical_type, Ordinal)
    assert column_logical_type.order == ['a', 'b', 'c']

    schema_df = pd.DataFrame(sample_series)
    schema_df.ww.init()

    schema_df.ww.set_types(logical_types={'sample_series': ordinal_with_order})
    logical_type = schema_df.ww.logical_types['sample_series']
    assert isinstance(logical_type, Ordinal)
    assert logical_type.order == ['a', 'b', 'c']


def test_ordinal_with_incomplete_ranking(sample_series):
    if _is_koalas_series(sample_series) or _is_dask_series(sample_series):
        pytest.xfail('Fails with Dask and Koalas - ordinal data validation not supported')

    ordinal_incomplete_order = Ordinal(order=['a', 'b'])
    error_msg = re.escape("Ordinal column sample_series contains values that are not "
                          "present in the order values provided: ['c']")

    schema_df = pd.DataFrame(sample_series)

    with pytest.raises(ValueError, match=error_msg):
        schema_df.ww.init(logical_types={'sample_series': ordinal_incomplete_order})

    schema_df.ww.init()
    with pytest.raises(ValueError, match=error_msg):
        schema_df.ww.set_types(logical_types={'sample_series': ordinal_incomplete_order})


def test_ordinal_with_nan_values():
    nan_df = pd.DataFrame(pd.Series(['a', 'b', np.nan, 'a'], name='nan_series'))
    ordinal_with_order = Ordinal(order=['a', 'b'])
    nan_df.ww.init(logical_types={'nan_series': ordinal_with_order})

    column_logical_type = nan_df.ww.logical_types['nan_series']
    assert isinstance(column_logical_type, Ordinal)
    assert column_logical_type.order == ['a', 'b']


def test_accessor_with_falsy_column_names(falsy_names_df):
    if _is_dask_dataframe(falsy_names_df):
        pytest.xfail('Dask DataFrames cannot handle integer column names')

    schema_df = falsy_names_df.copy()
    schema_df.ww.init(index=0, time_index='')
    assert schema_df.ww.index == 0
    assert schema_df.ww.time_index == ''

    schema_df.ww.set_time_index(None)
    assert schema_df.ww.time_index is None

    schema_df.ww.set_time_index('')
    assert schema_df.ww.time_index == ''

    popped_col = schema_df.ww.pop('')
    assert '' not in schema_df
    assert '' not in schema_df.ww.columns
    assert schema_df.ww.time_index is None

    schema_df.ww.set_index(None)
    assert schema_df.ww.index is None

    schema_df.ww[''] = popped_col
    assert schema_df.ww[''].name == ''

    renamed_df = schema_df.ww.rename({0: 'col_with_name'})
    assert 0 not in renamed_df.columns
    assert 'col_with_name' in renamed_df.columns


def test_dataframe_methods_on_accessor(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(name='test_schema')

    copied_df = schema_df.ww.copy()

    assert schema_df is not copied_df
    assert schema_df.ww._schema is not copied_df.ww._schema
    assert copied_df.ww.schema == schema_df.ww.schema

    pd.testing.assert_frame_equal(to_pandas(schema_df), to_pandas(copied_df))

    ltype_dtype = 'int64'
    new_dtype = 'string'

    warning = 'Operation performed by astype has invalidated the Woodwork typing information:\n '\
        f'dtype mismatch for column id between DataFrame dtype, {new_dtype}, and Integer dtype, {ltype_dtype}.\n '\
        'Please initialize Woodwork with DataFrame.ww.init'
    with pytest.warns(TypingInfoMismatchWarning, match=warning):
        new_df = schema_df.ww.astype({'id': new_dtype})
    assert new_df['id'].dtype == new_dtype
    assert new_df.ww.schema is None
    assert schema_df.ww.schema is not None


def test_dataframe_methods_on_accessor_new_schema_object(sample_df):
    sample_df.ww.init(index='id', semantic_tags={'email': 'new_tag'},
                      table_metadata={'contributors': ['user1', 'user2'],
                                      'created_on': '2/12/20'},
                      column_metadata={'id': {'important_keys': [1, 2, 3]}})

    copied_df = sample_df.ww.copy()

    assert sample_df.ww.schema == copied_df.ww.schema
    assert sample_df.ww._schema is not copied_df.ww._schema

    copied_df.ww.metadata['contributors'].append('user3')
    assert copied_df.ww.metadata == {'contributors': ['user1', 'user2', 'user3'],
                                     'created_on': '2/12/20'}

    assert sample_df.ww.metadata == {'contributors': ['user1', 'user2'],
                                     'created_on': '2/12/20'}

    copied_df.ww.reset_semantic_tags(retain_index_tags=False)
    assert copied_df.ww.index is None
    assert sample_df.ww.index == 'id'

    assert copied_df.ww.semantic_tags['email'] == set()
    assert sample_df.ww.semantic_tags['email'] == {'new_tag'}

    copied_df.ww.columns['id'].metadata['important_keys'].append(4)
    assert copied_df.ww.columns['id'].metadata == {'important_keys': [1, 2, 3, 4]}
    assert sample_df.ww.columns['id'].metadata == {'important_keys': [1, 2, 3]}


def test_dataframe_methods_on_accessor_inplace(sample_df):
    # TODO: Try to find a supported inplace method for Dask, if one exists
    if _is_dask_dataframe(sample_df):
        pytest.xfail('Dask does not support sort_values or rename inplace.')
    schema_df = sample_df.copy()
    schema_df.ww.init(name='test_schema')

    df_pre_sort = schema_df.copy()

    schema_df.ww.sort_values(['full_name'], inplace=True)
    assert schema_df.ww.name == 'test_schema'

    pd.testing.assert_frame_equal(to_pandas(schema_df), to_pandas(df_pre_sort.sort_values(['full_name'])))

    warning = "Operation performed by insert has invalidated the Woodwork typing information:\n "
    "The following columns in the DataFrame were missing from the typing information: {'new_name'}.\n "
    "Please initialize Woodwork with DataFrame.ww.init"
    with pytest.warns(TypingInfoMismatchWarning, match=warning):
        schema_df.ww.insert(loc=0, column="new_name", value=[1, 2, 3, 4])
    assert 'new_name' in schema_df.columns
    assert schema_df.ww.schema is None


def test_dataframe_methods_on_accessor_returning_series(sample_df):
    schema_df = sample_df[['id', 'age', 'is_registered']]
    schema_df.ww.init(name='test_schema')

    dtypes = schema_df.ww.dtypes

    assert schema_df.ww.name == 'test_schema'
    pd.testing.assert_series_equal(dtypes, schema_df.dtypes)
    all_series = schema_df.ww.all()
    assert schema_df.ww.name == 'test_schema'
    pd.testing.assert_series_equal(to_pandas(all_series), to_pandas(schema_df.all()))


def test_dataframe_methods_on_accessor_other_returns(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(name='test_schema')

    shape = schema_df.ww.shape

    assert schema_df.ww.name == 'test_schema'
    if _is_dask_dataframe(sample_df):
        shape = (shape[0].compute(), shape[1])
    assert shape == to_pandas(schema_df).shape
    assert schema_df.ww.name == 'test_schema'

    if not _is_dask_dataframe(sample_df):
        # keys() not supported with Dask
        pd.testing.assert_index_equal(schema_df.ww.keys(), schema_df.keys())


def test_dataframe_methods_on_accessor_to_pandas(sample_df):
    if isinstance(sample_df, pd.DataFrame):
        pytest.skip("No need to test converting pandas DataFrame to pandas")

    sample_df.ww.init(name='woodwork', index='id')

    if _is_dask_dataframe(sample_df):
        pd_df = sample_df.ww.compute()
    elif _is_koalas_dataframe(sample_df):
        pd_df = sample_df.ww.to_pandas()
        pytest.skip('Bug #1071: Woodwork not initialized after to_pandas call with Koalas categorical column')
    assert isinstance(pd_df, pd.DataFrame)
    assert pd_df.ww.index == 'id'
    assert pd_df.ww.name == 'woodwork'


def test_get_subset_df_with_schema(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(time_index='signup_date',
                      index='id',
                      name='df_name',
                      logical_types={'full_name': PersonFullName,
                                     'email': EmailAddress,
                                     'phone_number': PhoneNumber,
                                     'age': Double,
                                     'signup_date': Datetime},
                      semantic_tags={'full_name': ['new_tag', 'tag2'],
                                     'age': 'numeric'}
                      )
    schema = schema_df.ww.schema

    empty_df = schema_df.ww._get_subset_df_with_schema([])
    assert len(empty_df.columns) == 0
    assert empty_df.ww.schema is not None
    pd.testing.assert_frame_equal(to_pandas(empty_df), to_pandas(schema_df[[]]))
    validate_subset_schema(empty_df.ww.schema, schema)

    just_index = schema_df.ww._get_subset_df_with_schema(['id'])
    assert just_index.ww.index == schema.index
    assert just_index.ww.time_index is None
    pd.testing.assert_frame_equal(to_pandas(just_index), to_pandas(schema_df[['id']]))
    validate_subset_schema(just_index.ww.schema, schema)

    just_time_index = schema_df.ww._get_subset_df_with_schema(['signup_date'])
    assert just_time_index.ww.time_index == schema.time_index
    assert just_time_index.ww.index is None
    pd.testing.assert_frame_equal(to_pandas(just_time_index), to_pandas(schema_df[['signup_date']]))
    validate_subset_schema(just_time_index.ww.schema, schema)

    transfer_schema = schema_df.ww._get_subset_df_with_schema(['phone_number'])
    assert transfer_schema.ww.index is None
    assert transfer_schema.ww.time_index is None
    pd.testing.assert_frame_equal(to_pandas(transfer_schema), to_pandas(schema_df[['phone_number']]))
    validate_subset_schema(transfer_schema.ww.schema, schema)


def test_select_ltypes_no_match_and_all(sample_df, sample_correct_logical_types):
    schema_df = sample_df.copy()
    schema_df.ww.init(logical_types=sample_correct_logical_types)

    assert len(schema_df.ww.select(PostalCode).columns) == 0
    assert len(schema_df.ww.select(['PostalCode', PhoneNumber]).columns) == 1

    all_types = ww.type_system.registered_types
    assert len(schema_df.ww.select(exclude=all_types).columns) == 0

    df_all_types = schema_df.ww.select(all_types)
    pd.testing.assert_frame_equal(to_pandas(df_all_types), to_pandas(schema_df))
    assert df_all_types.ww.schema == schema_df.ww.schema


def test_select_ltypes_strings(sample_df, sample_correct_logical_types):
    schema_df = sample_df.copy()
    schema_df.ww.init(logical_types=sample_correct_logical_types)

    df_multiple_ltypes = schema_df.ww.select(['PersonFullName', 'email_address', 'double', 'BooleanNullable', 'datetime'])
    assert len(df_multiple_ltypes.columns) == 7
    assert 'phone_number' not in df_multiple_ltypes.columns
    assert 'id' not in df_multiple_ltypes.columns

    df_single_ltype = schema_df.ww.select('person_full_name')
    assert set(df_single_ltype.columns) == {'full_name'}


def test_select_ltypes_objects(sample_df, sample_correct_logical_types):
    schema_df = sample_df.copy()
    schema_df.ww.init(logical_types=sample_correct_logical_types)

    df_multiple_ltypes = schema_df.ww.select([PersonFullName, EmailAddress, Double, BooleanNullable, Datetime])
    assert len(df_multiple_ltypes.columns) == 7
    assert 'phone_number' not in df_multiple_ltypes.columns
    assert 'id' not in df_multiple_ltypes.columns

    df_single_ltype = schema_df.ww.select(PersonFullName)
    assert len(df_single_ltype.columns) == 1


def test_select_ltypes_mixed(sample_df, sample_correct_logical_types):
    schema_df = sample_df.copy()
    schema_df.ww.init(logical_types=sample_correct_logical_types)

    df_mixed_ltypes = schema_df.ww.select(['PersonFullName', 'email_address', Double])
    assert len(df_mixed_ltypes.columns) == 4
    assert 'phone_number' not in df_mixed_ltypes.columns


def test_select_ltypes_mixed_exclude(sample_df, sample_correct_logical_types):
    schema_df = sample_df.copy()
    schema_df.ww.init(logical_types=sample_correct_logical_types)

    df_mixed_ltypes = schema_df.ww.select(exclude=['PersonFullName', 'email_address', Double])
    assert len(df_mixed_ltypes.columns) == 10
    assert 'full_name' not in df_mixed_ltypes.columns
    assert 'email_address' not in df_mixed_ltypes.columns
    assert 'double' not in df_mixed_ltypes.columns
    assert 'double_with_nan' not in df_mixed_ltypes.columns


def test_select_ltypes_table(sample_df, sample_correct_logical_types):
    schema_df = sample_df.copy()
    schema_df.ww.init(name='testing',
                      index='id',
                      time_index='signup_date',
                      logical_types=sample_correct_logical_types,
                      semantic_tags={'full_name': ['new_tag', 'tag2'],
                                     'age': 'numeric',
                                     })

    df_no_indices = schema_df.ww.select('phone_number')
    assert df_no_indices.ww.index is None
    assert df_no_indices.ww.time_index is None

    df_with_indices = schema_df.ww.select(['Datetime', 'Integer'])
    assert df_with_indices.ww.index == 'id'
    assert df_with_indices.ww.time_index == 'signup_date'

    df_values = schema_df.ww.select(['PersonFullName'])
    assert df_values.ww.name == schema_df.ww.name
    assert df_values.ww.columns['full_name'] == schema_df.ww.columns['full_name']


def test_select_semantic_tags(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(semantic_tags={'full_name': 'tag1',
                                     'email': ['tag2'],
                                     'age': ['numeric', 'tag2'],
                                     'phone_number': ['tag3', 'tag2'],
                                     'is_registered': 'category',
                                     },
                      time_index='signup_date')

    df_one_match = schema_df.ww.select('numeric')
    assert len(df_one_match.columns) == 6
    assert 'age' in df_one_match.columns
    assert 'id' in df_one_match.columns

    df_multiple_matches = schema_df.ww.select('tag2')
    assert len(df_multiple_matches.columns) == 3
    assert 'age' in df_multiple_matches.columns
    assert 'phone_number' in df_multiple_matches.columns
    assert 'email' in df_multiple_matches.columns

    df_multiple_tags = schema_df.ww.select(['numeric', 'time_index'])
    assert len(df_multiple_tags.columns) == 7
    assert 'id' in df_multiple_tags.columns
    assert 'age' in df_multiple_tags.columns
    assert 'signup_date' in df_multiple_tags.columns

    df_overlapping_tags = schema_df.ww.select(['numeric', 'tag2'])
    assert len(df_overlapping_tags.columns) == 8
    assert 'id' in df_overlapping_tags.columns
    assert 'age' in df_overlapping_tags.columns
    assert 'phone_number' in df_overlapping_tags.columns
    assert 'email' in df_overlapping_tags.columns

    df_common_tags = schema_df.ww.select(['category', 'numeric'])
    assert len(df_common_tags.columns) == 8
    assert 'id' in df_common_tags.columns
    assert 'is_registered' in df_common_tags.columns
    assert 'age' in df_common_tags.columns


def test_select_semantic_tags_exclude(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(semantic_tags={'full_name': 'tag1',
                                     'email': ['tag2'],
                                     'age': ['numeric', 'tag2'],
                                     'phone_number': ['tag3', 'tag2'],
                                     'is_registered': 'category',
                                     },
                      time_index='signup_date')

    df_one_match = schema_df.ww.select(exclude='numeric')
    assert len(df_one_match.columns) == 8
    assert 'age' not in df_one_match.columns
    assert 'id' not in df_one_match.columns

    df_multiple_matches = schema_df.ww.select(exclude='tag2')
    assert len(df_multiple_matches.columns) == 11
    assert 'age' not in df_multiple_matches.columns
    assert 'phone_number' not in df_multiple_matches.columns
    assert 'email' not in df_multiple_matches.columns

    df_multiple_tags = schema_df.ww.select(exclude=['numeric', 'time_index'])
    assert len(df_multiple_tags.columns) == 7
    assert 'id' not in df_multiple_tags.columns
    assert 'age' not in df_multiple_tags.columns
    assert 'signup_date' not in df_multiple_tags.columns

    df_overlapping_tags = schema_df.ww.select(exclude=['numeric', 'tag2'])
    assert len(df_overlapping_tags.columns) == 6
    assert 'id' not in df_overlapping_tags.columns
    assert 'age' not in df_overlapping_tags.columns
    assert 'phone_number' not in df_overlapping_tags.columns
    assert 'email' not in df_overlapping_tags.columns

    df_common_tags = schema_df.ww.select(exclude=['category', 'numeric'])
    assert len(df_common_tags.columns) == 6
    assert 'id' not in df_common_tags.columns
    assert 'is_registered' not in df_common_tags.columns
    assert 'age' not in df_common_tags.columns


def test_select_single_inputs(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(time_index='signup_date',
                      index='id',
                      name='df_name',
                      logical_types={
                          'full_name': PersonFullName,
                          'email': EmailAddress,
                          'phone_number': PhoneNumber,
                          'signup_date': Datetime(datetime_format='%Y-%m-%d')
                      },
                      semantic_tags={
                          'full_name': ['new_tag', 'tag2'],
                          'age': 'numeric',
                          'signup_date': 'date_of_birth'
                      })

    df_ltype_string = schema_df.ww.select('person_full_name')
    assert len(df_ltype_string.columns) == 1
    assert 'full_name' in df_ltype_string.columns

    df_ltype_obj = schema_df.ww.select(IntegerNullable)
    assert len(df_ltype_obj.columns) == 2
    assert 'age' in df_ltype_obj.columns

    df_tag_string = schema_df.ww.select('index')
    assert len(df_tag_string.columns) == 1
    assert 'id' in df_tag_string.columns

    df_tag_instantiated = schema_df.ww.select('Datetime')
    assert len(df_tag_instantiated.columns) == 2
    assert 'signup_date' in df_tag_instantiated.columns


def test_select_list_inputs(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(time_index='signup_date',
                      index='id',
                      name='df_name',
                      logical_types={
                          'full_name': PersonFullName,
                          'email': EmailAddress,
                          'phone_number': PhoneNumber,
                          'signup_date': Datetime(datetime_format='%Y-%m-%d'),
                      },
                      semantic_tags={
                          'full_name': ['new_tag', 'tag2'],
                          'age': 'numeric',
                          'signup_date': 'date_of_birth',
                          'email': 'tag2',
                          'is_registered': 'category'
                      })
    df_just_strings = schema_df.ww.select(['PersonFullName', 'index', 'tag2', 'boolean_nullable'])
    assert len(df_just_strings.columns) == 4
    assert 'id' in df_just_strings.columns
    assert 'full_name' in df_just_strings.columns
    assert 'email' in df_just_strings.columns
    assert 'is_registered' in df_just_strings.columns

    df_mixed_selectors = schema_df.ww.select([PersonFullName, 'index', 'time_index', Integer])
    assert len(df_mixed_selectors.columns) == 4
    assert 'id' in df_mixed_selectors.columns
    assert 'full_name' in df_mixed_selectors.columns
    assert 'signup_date' in df_mixed_selectors.columns

    df_common_tags = schema_df.ww.select(['category', 'numeric', BooleanNullable, Datetime])
    assert len(df_common_tags.columns) == 9
    assert 'is_registered' in df_common_tags.columns
    assert 'age' in df_common_tags.columns
    assert 'signup_date' in df_common_tags.columns


def test_select_semantic_tags_no_match(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(time_index='signup_date',
                      index='id',
                      name='df_name',
                      logical_types={
                          'full_name': PersonFullName,
                          'email': EmailAddress,
                          'phone_number': PhoneNumber,
                          'signup_date': Datetime(datetime_format='%Y-%m-%d'),
                      }, semantic_tags={
                          'full_name': ['new_tag', 'tag2'],
                          'age': 'numeric',
                          'signup_date': 'date_of_birth',
                          'email': 'tag2'
                      })

    assert len(schema_df.ww.select(['doesnt_exist']).columns) == 0
    df_multiple_unused = schema_df.ww.select(['doesnt_exist', 'boolean_nullable', 'category', PhoneNumber])
    assert len(df_multiple_unused.columns) == 3

    df_unused_ltype = schema_df.ww.select(['date_of_birth', 'doesnt_exist', PostalCode, Integer])
    assert len(df_unused_ltype.columns) == 3


def test_select_repetitive(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(time_index='signup_date',
                      index='id',
                      name='df_name',
                      logical_types={
                          'full_name': PersonFullName,
                          'email': EmailAddress,
                          'phone_number': PhoneNumber,
                          'signup_date': Datetime(datetime_format='%Y-%m-%d'),
                      }, semantic_tags={
                          'full_name': ['new_tag', 'tag2'],
                          'age': 'numeric',
                          'signup_date': 'date_of_birth',
                          'email': 'tag2'
                      })
    df_repeat_tags = schema_df.ww.select(['new_tag', 'new_tag'])
    assert len(df_repeat_tags.columns) == 1
    assert set(df_repeat_tags.columns) == {'full_name'}

    df_repeat_ltypes = schema_df.ww.select(['PhoneNumber', PhoneNumber, 'phone_number'])
    assert len(df_repeat_ltypes.columns) == 1
    assert set(df_repeat_ltypes.columns) == {'phone_number'}


def test_select_instantiated_ltype():
    ymd_format = Datetime(datetime_format='%Y~%m~%d')

    df = pd.DataFrame({
        'dates': ["2019/01/01", "2019/01/02", "2019/01/03"],
        'ymd': ["2019~01~01", "2019~01~02", "2019~01~03"],
    })
    df.ww.init(
        logical_types={'ymd': ymd_format,
                       'dates': Datetime})

    new_df = df.ww.select('Datetime')
    assert len(new_df.columns) == 2

    new_df = df.ww.select(Datetime)
    assert len(new_df.columns) == 2

    err_msg = "Invalid selector used in include: Datetime cannot be instantiated"
    with pytest.raises(TypeError, match=err_msg):
        df.ww.select(ymd_format)


def test_select_return_schema(sample_df):
    sample_df.ww.init()
    # Multiple column matches
    df_schema = sample_df.ww.select(include='Unknown', return_schema=True)
    assert isinstance(df_schema, TableSchema)
    assert len(df_schema.columns) == 2
    assert df_schema == sample_df.ww.select(include='Unknown').ww.schema

    # Single column match
    single_schema = sample_df.ww.select(include='BooleanNullable', return_schema=True)
    assert isinstance(single_schema, TableSchema)
    assert len(single_schema.columns) == 1
    assert single_schema == sample_df.ww.select(include='BooleanNullable').ww.schema

    # No matches
    empty_schema = sample_df.ww.select(include='IPAddress', return_schema=True)
    assert isinstance(empty_schema, TableSchema)
    assert len(empty_schema.columns) == 0


@pytest.mark.parametrize(
    "ww_type, pandas_type",
    [(["Integer", "IntegerNullable"], "int"),
     (["Double"], "float"),
     (["Datetime"], "datetime"),
     (["Unknown", "EmailAddress"], "string"),
     (["Categorical"], "category"),
     (["Boolean", "BooleanNullable"], "boolean")]
)
def test_select_retains_column_order(ww_type, pandas_type, sample_df):
    if _is_koalas_dataframe(sample_df) and pandas_type in ["category", "string"]:
        pytest.skip("Koalas stores categories as strings")
    sample_df.ww.init()

    ww_schema_column_order = [x for x in sample_df.ww.select(ww_type, return_schema=True).columns.keys()]
    pandas_column_order = [x for x in sample_df.select_dtypes(include=pandas_type).columns]
    assert ww_schema_column_order == pandas_column_order


def test_select_include_and_exclude_error(sample_df):
    sample_df.ww.init()
    err_msg = "Cannot specify values for both 'include' and 'exclude' in a single call."
    with pytest.raises(ValueError, match=err_msg):
        sample_df.ww.select(include='Integer', exclude='Double')

    with pytest.raises(ValueError, match=err_msg):
        sample_df.ww.select(include=[], exclude=[])


def test_select_no_selectors_error(sample_df):
    sample_df.ww.init()
    err_msg = "Must specify values for either 'include' or 'exclude'."
    with pytest.raises(ValueError, match=err_msg):
        sample_df.ww.select()


def test_accessor_set_index(sample_df):
    sample_df.ww.init()

    sample_df.ww.set_index('id')
    assert sample_df.ww.index == 'id'
    if isinstance(sample_df, pd.DataFrame):
        # underlying index not set for Dask/Koalas
        assert (sample_df.index == sample_df['id']).all()

    sample_df.ww.set_index('full_name')
    assert sample_df.ww.index == 'full_name'
    if isinstance(sample_df, pd.DataFrame):
        # underlying index not set for Dask/Koalas
        assert (sample_df.index == sample_df['full_name']).all()

    sample_df.ww.set_index(None)
    assert sample_df.ww.index is None
    if isinstance(sample_df, pd.DataFrame):
        # underlying index not set for Dask/Koalas
        # Check that underlying index doesn't get reset when Woodwork index is removed
        assert (sample_df.index == sample_df['full_name']).all()


def test_accessor_set_index_errors(sample_df):
    sample_df.ww.init()

    error = 'Specified index column `testing` not found in TableSchema.'
    with pytest.raises(ColumnNotPresentError, match=error):
        sample_df.ww.set_index('testing')

    if isinstance(sample_df, pd.DataFrame):
        # Index uniqueness not validate for Dask/Koalas
        error = "Index column must be unique"
        with pytest.raises(LookupError, match=error):
            sample_df.ww.set_index('age')


def test_set_types(sample_df):
    sample_df.ww.init(index='full_name', time_index='signup_date')

    original_df = sample_df.ww.copy()

    sample_df.ww.set_types()
    assert original_df.ww.schema == sample_df.ww.schema
    pd.testing.assert_frame_equal(to_pandas(original_df), to_pandas(sample_df))

    sample_df.ww.set_types(logical_types={'is_registered': 'IntegerNullable'})
    assert sample_df['is_registered'].dtype == 'Int64'

    sample_df.ww.set_types(semantic_tags={'signup_date': ['new_tag']},
                           logical_types={'full_name': 'Categorical'},
                           retain_index_tags=False)
    assert sample_df.ww.index is None
    assert sample_df.ww.time_index is None


def test_set_types_errors(sample_df):
    sample_df.ww.init(index='full_name')

    error = "String invalid is not a valid logical type"
    with pytest.raises(ValueError, match=error):
        sample_df.ww.set_types(logical_types={'id': 'invalid'})

    if isinstance(sample_df, pd.DataFrame):
        # Dask does not error on invalid type conversion until compute
        # Koalas does conversion and fills values with NaN
        error = 'Error converting datatype for email from type string ' \
            'to type float64. Please confirm the underlying data is consistent with ' \
            'logical type Double.'
        with pytest.raises(TypeConversionError, match=error):
            sample_df.ww.set_types(logical_types={'email': 'Double'})

    error = re.escape("Cannot add 'index' tag directly for column email. To set a column as the index, "
                      "use DataFrame.ww.set_index() instead.")
    with pytest.raises(ValueError, match=error):
        sample_df.ww.set_types(semantic_tags={'email': 'index'})


def test_pop(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init(semantic_tags={'age': 'custom_tag'})
    original_schema = schema_df.ww.schema

    popped_series = schema_df.ww.pop('age')

    assert isinstance(popped_series, type(sample_df['age']))
    assert popped_series.ww.semantic_tags == {'custom_tag', 'numeric'}
    pd.testing.assert_series_equal(to_pandas(popped_series), pd.Series([pd.NA, 33, 33, 57], dtype='Int64', name='age'))
    assert isinstance(popped_series.ww.logical_type, IntegerNullable)

    assert 'age' not in schema_df.columns
    assert 'age' not in schema_df.ww.columns

    assert 'age' not in schema_df.ww.logical_types.keys()
    assert 'age' not in schema_df.ww.semantic_tags.keys()

    assert schema_df.ww.schema == original_schema._get_subset_schema(list(schema_df.columns))

    schema_df = sample_df.copy()
    schema_df.ww.init(
        name='table',
        logical_types={'age': IntegerNullable},
        semantic_tags={'age': 'custom_tag'},
        use_standard_tags=False)

    popped_series = schema_df.ww.pop('age')

    assert popped_series.ww.semantic_tags == {'custom_tag'}


def test_pop_index(sample_df):
    sample_df.ww.init(index='id', name='df_name')
    assert sample_df.ww.index == 'id'
    id_col = sample_df.ww.pop('id')
    assert sample_df.ww.index is None
    assert 'index' in id_col.ww.semantic_tags


def test_pop_error(sample_df):
    sample_df.ww.init(
        name='table',
        logical_types={'age': IntegerNullable},
        semantic_tags={'age': 'custom_tag'},
        use_standard_tags=True)

    with pytest.raises(ColumnNotPresentError, match="Column with name 'missing' not found in DataFrame"):
        sample_df.ww.pop("missing")


def test_accessor_drop(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init()

    single_input_df = schema_df.ww.drop('is_registered')
    assert len(single_input_df.ww.columns) == (len(schema_df.columns) - 1)
    assert 'is_registered' not in single_input_df.ww.columns
    assert to_pandas(schema_df).drop('is_registered', axis='columns').equals(to_pandas(single_input_df))

    list_input_df = schema_df.ww.drop(['is_registered'])
    assert len(list_input_df.ww.columns) == (len(schema_df.columns) - 1)
    assert 'is_registered' not in list_input_df.ww.columns
    assert to_pandas(schema_df).drop('is_registered', axis='columns').equals(to_pandas(list_input_df))
    # should be equal to the single input example above
    assert single_input_df.ww.schema == list_input_df.ww.schema
    assert to_pandas(single_input_df).equals(to_pandas(list_input_df))

    multiple_list_df = schema_df.ww.drop(['age', 'full_name', 'is_registered'])
    assert len(multiple_list_df.ww.columns) == (len(schema_df.columns) - 3)
    assert 'is_registered' not in multiple_list_df.ww.columns
    assert 'full_name' not in multiple_list_df.ww.columns
    assert 'age' not in multiple_list_df.ww.columns

    assert to_pandas(schema_df).drop(['is_registered', 'age', 'full_name'], axis='columns').equals(to_pandas(multiple_list_df))

    # Drop the same columns in a different order and confirm resulting DataFrame column order doesn't change
    different_order_df = schema_df.ww.drop(['is_registered', 'age', 'full_name'])
    assert different_order_df.ww.schema == multiple_list_df.ww.schema
    assert to_pandas(multiple_list_df).equals(to_pandas(different_order_df))


def test_accessor_drop_inplace(sample_df):
    sample_df.ww.init()
    inplace_df = sample_df.copy()
    inplace_df.ww.init()

    if _is_dask_dataframe(sample_df):
        error = 'Drop inplace not supported for Dask'
        with pytest.raises(ValueError, match=error):
            inplace_df.ww.drop(['is_registered'], inplace=True)
    elif _is_koalas_dataframe(sample_df):
        error = 'Drop inplace not supported for Koalas'
        with pytest.raises(ValueError, match=error):
            inplace_df.ww.drop(['is_registered'], inplace=True)
    else:
        inplace_df.ww.drop(['is_registered'], inplace=True)
        assert len(inplace_df.ww.columns) == (len(sample_df.columns) - 1)
        assert 'is_registered' not in inplace_df.ww.columns

        assert sample_df.drop('is_registered', axis='columns').equals(inplace_df)


def test_accessor_drop_indices(sample_df):
    sample_df.ww.init(index='id', time_index='signup_date')
    assert sample_df.ww.index == 'id'
    assert sample_df.ww.time_index == 'signup_date'

    dropped_index_df = sample_df.ww.drop('id')
    assert 'id' not in dropped_index_df.ww.columns
    assert dropped_index_df.ww.index is None
    assert dropped_index_df.ww.time_index == 'signup_date'

    dropped_time_index_df = sample_df.ww.drop(['signup_date'])
    assert 'signup_date' not in dropped_time_index_df.ww.columns
    assert dropped_time_index_df.ww.time_index is None
    assert dropped_time_index_df.ww.index == 'id'


def test_accessor_drop_errors(sample_df):
    sample_df.ww.init()

    error = re.escape("Column(s) '['not_present']' not found in DataFrame")
    with pytest.raises(ColumnNotPresentError, match=error):
        sample_df.ww.drop('not_present')

    with pytest.raises(ColumnNotPresentError, match=error):
        sample_df.ww.drop(['age', 'not_present'])

    error = re.escape("Column(s) '['not_present1', 4]' not found in DataFrame")
    with pytest.raises(ColumnNotPresentError, match=error):
        sample_df.ww.drop(['not_present1', 4])


def test_accessor_rename(sample_df):
    table_metadata = {'table_info': 'this is text'}
    id_description = 'the id of the row'
    id_origin = 'base'
    sample_df.ww.init(index='id',
                      time_index='signup_date',
                      table_metadata=table_metadata,
                      column_descriptions={'id': id_description},
                      column_origins={'id': id_origin},
                      semantic_tags={'age': 'test_tag'},
                      logical_types={'age': Double})
    original_df = sample_df.ww.copy()

    new_df = sample_df.ww.rename({'age': 'birthday'})

    assert to_pandas(sample_df.rename(columns={'age': 'birthday'})).equals(to_pandas(new_df))
    # Confirm original dataframe hasn't changed
    assert to_pandas(sample_df).equals(to_pandas(original_df))
    assert sample_df.ww.schema == original_df.ww.schema

    assert original_df.columns.get_loc('age') == new_df.columns.get_loc('birthday')
    pd.testing.assert_series_equal(to_pandas(original_df['age']),
                                   to_pandas(new_df['birthday']),
                                   check_names=False)

    # confirm that metadata and descriptions are there
    assert new_df.ww.metadata == table_metadata
    assert new_df.ww.columns['id'].description == id_description
    assert new_df.ww.columns['id'].origin == id_origin

    old_col = sample_df.ww.columns['age']
    new_col = new_df.ww.columns['birthday']
    assert old_col.logical_type == new_col.logical_type
    assert old_col.semantic_tags == new_col.semantic_tags

    new_df = sample_df.ww.rename({'age': 'full_name', 'full_name': 'age'})

    pd.testing.assert_series_equal(to_pandas(original_df['age']),
                                   to_pandas(new_df['full_name']),
                                   check_names=False)
    pd.testing.assert_series_equal(to_pandas(original_df['full_name']),
                                   to_pandas(new_df['age']),
                                   check_names=False)

    assert original_df.columns.get_loc('age') == new_df.columns.get_loc('full_name')
    assert original_df.columns.get_loc('full_name') == new_df.columns.get_loc('age')


def test_accessor_rename_inplace(sample_df):
    table_metadata = {'table_info': 'this is text'}
    id_description = 'the id of the row'
    id_origin = 'base'
    sample_df.ww.init(index='id',
                      time_index='signup_date',
                      table_metadata=table_metadata,
                      column_descriptions={'id': id_description},
                      column_origins={'id': id_origin},
                      semantic_tags={'age': 'test_tag'},
                      logical_types={'age': Double})
    original_df = sample_df.ww.copy()
    inplace_df = sample_df.ww.copy()

    if _is_dask_dataframe(sample_df):
        error = 'Rename inplace not supported for Dask'
        with pytest.raises(ValueError, match=error):
            inplace_df.ww.rename({'age': 'birthday'}, inplace=True)
    elif _is_koalas_dataframe(sample_df):
        error = 'Rename inplace not supported for Koalas'
        with pytest.raises(ValueError, match=error):
            inplace_df.ww.rename({'age': 'birthday'}, inplace=True)

    else:
        inplace_df.ww.rename({'age': 'birthday'}, inplace=True)

        assert original_df.columns.get_loc('age') == inplace_df.columns.get_loc('birthday')
        pd.testing.assert_series_equal(to_pandas(original_df['age']),
                                       to_pandas(inplace_df['birthday']),
                                       check_names=False)

        # confirm that metadata and descriptions are there
        assert inplace_df.ww.metadata == table_metadata
        assert inplace_df.ww.columns['id'].description == id_description
        assert inplace_df.ww.columns['id'].origin == id_origin

        old_col = sample_df.ww.columns['age']
        new_col = inplace_df.ww.columns['birthday']
        assert old_col.logical_type == new_col.logical_type
        assert old_col.semantic_tags == new_col.semantic_tags

        new_df = sample_df.ww.copy()
        new_df.ww.rename({'age': 'full_name', 'full_name': 'age'}, inplace=True)

        pd.testing.assert_series_equal(to_pandas(original_df['age']),
                                       to_pandas(new_df['full_name']),
                                       check_names=False)
        pd.testing.assert_series_equal(to_pandas(original_df['full_name']),
                                       to_pandas(new_df['age']),
                                       check_names=False)

        assert original_df.columns.get_loc('age') == new_df.columns.get_loc('full_name')
        assert original_df.columns.get_loc('full_name') == new_df.columns.get_loc('age')


def test_accessor_rename_indices(sample_df):
    sample_df.ww.init(
        index='id',
        time_index='signup_date')

    renamed_df = sample_df.ww.rename({'id': 'renamed_index', 'signup_date': 'renamed_time_index'})
    assert 'id' not in renamed_df.columns
    assert 'signup_date' not in renamed_df.columns
    assert 'renamed_index' in renamed_df.columns
    assert 'renamed_time_index' in renamed_df.columns

    if isinstance(sample_df, pd.DataFrame):
        # underlying index not set for Dask/Koalas
        assert all(renamed_df.index == renamed_df['renamed_index'])

    assert renamed_df.ww.index == 'renamed_index'
    assert renamed_df.ww.time_index == 'renamed_time_index'


def test_accessor_schema_properties(sample_df):
    sample_df.ww.init(index='id',
                      time_index='signup_date')

    schema_properties = ['logical_types', 'semantic_tags', 'index', 'time_index', 'use_standard_tags']
    for schema_property in schema_properties:
        prop_from_accessor = getattr(sample_df.ww, schema_property)
        prop_from_schema = getattr(sample_df.ww.schema, schema_property)

        assert prop_from_accessor == prop_from_schema

        # Assumes we don't have setters for any of these attributes
        error = "can't set attribute"
        with pytest.raises(AttributeError, match=error):
            setattr(sample_df.ww, schema_property, 'new_value')


def test_sets_koalas_option_on_init(sample_df_koalas):
    if ks:
        ks.set_option('compute.ops_on_diff_frames', False)
        sample_df_koalas.ww.init()
        assert ks.get_option('compute.ops_on_diff_frames') is True


def test_setitem_invalid_input(sample_df):
    df = sample_df.copy()
    df.ww.init(index='id', time_index='signup_date')

    error_msg = 'New column must be of Series type'
    with pytest.raises(ValueError, match=error_msg):
        df.ww['test'] = [1, 2, 3]

    error_msg = 'Cannot reassign index. Change column name and then use df.ww.set_index to reassign index.'
    with pytest.raises(KeyError, match=error_msg):
        df.ww['id'] = df.id

    error_msg = 'Cannot reassign time index. Change column name and then use df.ww.set_time_index to reassign time index.'
    with pytest.raises(KeyError, match=error_msg):
        df.ww['signup_date'] = df.signup_date


def test_setitem_indexed_column_on_unindexed_dataframe(sample_df):
    sample_df.ww.init()

    col = sample_df.ww.pop('id')
    col.ww.add_semantic_tags(semantic_tags='index')

    warning = 'Cannot add "index" tag on id directly to the DataFrame. The "index" tag has been removed from id. To set this column as a Woodwork index, please use df.ww.set_index'

    with pytest.warns(IndexTagRemovedWarning, match=warning):
        sample_df.ww['id'] = col

    assert sample_df.ww.index is None
    assert ww.is_schema_valid(sample_df, sample_df.ww.schema)
    assert sample_df.ww['id'].ww.semantic_tags == {'numeric'}


def test_setitem_indexed_column_on_indexed_dataframe(sample_df):
    sample_df.ww.init()
    sample_df.ww.set_index('id')

    col = sample_df.ww.pop('id')

    warning = 'Cannot add "index" tag on id directly to the DataFrame. The "index" tag has been removed from id. To set this column as a Woodwork index, please use df.ww.set_index'

    with pytest.warns(IndexTagRemovedWarning, match=warning):
        sample_df.ww['id'] = col

    assert sample_df.ww.index is None
    assert ww.is_schema_valid(sample_df, sample_df.ww.schema)
    assert sample_df.ww['id'].ww.semantic_tags == {'numeric'}

    sample_df.ww.init(logical_types={'email': 'Categorical'})
    sample_df.ww.set_index('id')

    col = sample_df.ww.pop('email')
    col.ww.add_semantic_tags(semantic_tags='index')

    warning = 'Cannot add "index" tag on email directly to the DataFrame. The "index" tag has been removed from email. To set this column as a Woodwork index, please use df.ww.set_index'

    with pytest.warns(IndexTagRemovedWarning, match=warning):
        sample_df.ww['email'] = col
    assert sample_df.ww.index == 'id'
    assert sample_df.ww.semantic_tags['email'] == {'category'}


def test_setitem_indexed_column_on_unindexed_dataframe_no_standard_tags(sample_df):
    sample_df.ww.init()

    col = sample_df.ww.pop('id')
    col.ww.init(semantic_tags='index', use_standard_tags=False)

    warning = 'Cannot add "index" tag on id directly to the DataFrame. The "index" tag has been removed from id. To set this column as a Woodwork index, please use df.ww.set_index'

    with pytest.warns(IndexTagRemovedWarning, match=warning):
        sample_df.ww['id'] = col

    assert sample_df.ww.index is None
    assert ww.is_schema_valid(sample_df, sample_df.ww.schema)
    assert sample_df.ww['id'].ww.semantic_tags == set()


def test_setitem_different_name(sample_df):
    df = sample_df.copy()
    df.ww.init()

    new_series = pd.Series([1, 2, 3, 4], name='wrong', dtype='float')
    if _is_koalas_dataframe(sample_df):
        new_series = ks.Series(new_series)

    # Assign series with name `wrong` to existing column with name `id`
    df.ww['id'] = new_series
    assert df.ww['id'].name == 'id'
    assert 'id' in df.ww.columns
    assert 'wrong' not in df.ww.columns
    assert 'wrong' not in df.columns

    new_series2 = pd.Series([1, 2, 3, 4], name='wrong2', dtype='float')
    if _is_koalas_dataframe(sample_df):
        new_series2 = ks.Series(new_series2)

    # Assign series with name `wrong2` to new column with name `new_col`
    df.ww['new_col'] = new_series2
    assert df.ww['new_col'].name == 'new_col'
    assert 'new_col' in df.ww.columns
    assert 'wrong2' not in df.ww.columns
    assert 'wrong2' not in df.columns


def test_setitem_new_column(sample_df):
    df = sample_df.copy()
    df.ww.init(use_standard_tags=False)

    new_series = pd.Series([1, 2, 3, 4])
    if _is_koalas_dataframe(sample_df):
        new_series = ks.Series(new_series)
    dtype = 'int64'

    df.ww['test_col2'] = new_series
    assert 'test_col2' in df.columns
    assert 'test_col2' in df.ww._schema.columns.keys()
    assert isinstance(df.ww['test_col2'].ww.logical_type, Integer)
    assert df.ww['test_col2'].ww.use_standard_tags is True
    assert df.ww['test_col2'].ww.semantic_tags == {'numeric'}
    assert df.ww['test_col2'].name == 'test_col2'
    assert df.ww['test_col2'].dtype == dtype

    new_series = pd.Series([1, 2, 3], dtype='float')
    if _is_koalas_dataframe(sample_df):
        new_series = ks.Series(new_series)

    new_series = init_series(
        new_series,
        logical_type=Double,
        use_standard_tags=False,
        semantic_tags={'test_tag'},
    )

    df.ww['test_col3'] = new_series
    assert 'test_col3' in df.ww.columns
    assert isinstance(df.ww['test_col3'].ww.logical_type, Double)
    assert df.ww['test_col3'].ww.use_standard_tags is False
    assert df.ww['test_col3'].ww.semantic_tags == {'test_tag'}
    assert df.ww['test_col3'].name == 'test_col3'
    assert df.ww['test_col3'].dtype == 'float'

    # Standard tags and no logical type
    df = sample_df.copy()
    df.ww.init(use_standard_tags=True)

    new_series = pd.Series(['new', 'column', 'inserted'], name='test_col')
    if _is_koalas_dataframe(sample_df):
        dtype = 'string'
        new_series = ks.Series(new_series)
    else:
        dtype = 'category'

    new_series = init_series(new_series, logical_type="Categorical")
    df.ww['test_col'] = new_series
    assert 'test_col' in df.ww.columns
    assert isinstance(df.ww['test_col'].ww.logical_type, Categorical)
    assert df.ww['test_col'].ww.use_standard_tags is True
    assert df.ww['test_col'].ww.semantic_tags == {'category'}
    assert df.ww['test_col'].name == 'test_col'
    assert df.ww['test_col'].dtype == dtype


def test_setitem_overwrite_column(sample_df):
    df = sample_df.copy()
    df.ww.init(
        index='id',
        time_index='signup_date',
        use_standard_tags=True
    )

    # Change to column no change in types
    original_col = df.ww['age']
    new_series = pd.Series([1, 2, 3, None], dtype='Int64')
    if _is_koalas_dataframe(sample_df):
        new_series = ks.Series(new_series)

    dtype = 'Int64'
    new_series = init_series(new_series, use_standard_tags=True)
    df.ww['age'] = new_series

    assert 'age' in df.columns
    assert 'age' in df.ww._schema.columns.keys()
    assert isinstance(df.ww['age'].ww.logical_type, type(original_col.ww.logical_type))
    assert df.ww['age'].ww.semantic_tags == original_col.ww.semantic_tags
    assert df.ww['age'].dtype == dtype
    assert original_col is not df.ww['age']

    # Change dtype, logical types, and tags with conflicting use_standard_tags
    original_col = df['full_name']
    new_series = pd.Series([0, 1, 2], dtype='float')
    if _is_koalas_dataframe(sample_df):
        new_series = ks.Series(new_series)

    new_series = init_series(
        new_series,
        use_standard_tags=False,
        semantic_tags='test_tag',
    )

    df.ww['full_name'] = new_series

    assert 'full_name' in df.columns
    assert 'full_name' in df.ww._schema.columns.keys()
    assert isinstance(df.ww['full_name'].ww.logical_type, Double)
    assert df.ww['full_name'].ww.use_standard_tags is False
    assert df.ww['full_name'].ww.semantic_tags == {'test_tag'}
    assert df.ww['full_name'].dtype == 'float'
    assert original_col is not df.ww['full_name']

    df = sample_df.copy()
    df.ww.init(use_standard_tags=False)

    original_col = df['full_name']
    new_series = pd.Series([0, 1, 2], dtype='float')
    if _is_koalas_dataframe(sample_df):
        new_series = ks.Series(new_series)

    new_series = init_series(
        new_series,
        use_standard_tags=True,
        semantic_tags='test_tag',
    )

    df.ww['full_name'] = new_series

    assert 'full_name' in df.columns
    assert 'full_name' in df.ww._schema.columns.keys()
    assert isinstance(df.ww['full_name'].ww.logical_type, Double)
    assert df.ww['full_name'].ww.use_standard_tags is True
    assert df.ww['full_name'].ww.semantic_tags == {'test_tag', 'numeric'}
    assert df.ww['full_name'].dtype == 'float'
    assert original_col is not df.ww['full_name']


def test_maintain_column_order_on_type_changes(sample_df):
    sample_df.ww.init()
    schema_df = sample_df.ww.copy()

    schema_df.ww.set_types(logical_types={'email': 'Categorical', 'id': 'Double'},
                           semantic_tags={'age': 'tag', 'email': 'tag'})
    assert all(schema_df.columns == sample_df.columns)
    assert all(schema_df.ww.types.index == sample_df.ww.types.index)

    schema_df.ww.set_index('email')
    assert all(schema_df.columns == sample_df.columns)
    assert all(schema_df.ww.types.index == sample_df.ww.types.index)


def test_maintain_column_order_of_dataframe(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init()
    select_df = schema_df.ww.select([Unknown, EmailAddress, Integer, IntegerNullable, Boolean, BooleanNullable, Datetime, Double, Categorical])
    assert all(schema_df.columns == select_df.columns)
    assert all(schema_df.ww.types.index == select_df.ww.types.index)

    renamed_df = schema_df.ww.rename({'email': 'renamed_1', 'id': 'renamed_2'})
    renamed_cols = ['renamed_2', 'full_name', 'renamed_1', 'phone_number', 'age',
                    'signup_date', 'is_registered', 'double', 'double_with_nan', 'integer',
                    'nullable_integer', 'boolean', 'categorical', 'datetime_with_NaT']
    assert all(renamed_cols == renamed_df.columns)
    assert all(renamed_cols == renamed_df.ww.types.index)

    dropped_df = schema_df.ww.drop(['email', 'id', 'is_registered', 'age'])
    cols_left_over = ['full_name', 'phone_number', 'signup_date', 'double', 'double_with_nan',
                      'integer', 'nullable_integer', 'boolean', 'categorical',
                      'datetime_with_NaT']
    assert all(cols_left_over == dropped_df.columns)
    assert all(cols_left_over == dropped_df.ww.types.index)


def test_maintain_column_order_of_input(sample_df):
    schema_df = sample_df.copy()
    schema_df.ww.init()

    reversed_cols = list(schema_df.columns[::-1])

    if not _is_dask_dataframe(sample_df):
        iloc_df = schema_df.ww.iloc[:, list(range(len(schema_df.columns)))[::-1]]
        assert all(reversed_cols == iloc_df.columns)
        assert all(reversed_cols == iloc_df.ww.types.index)

    loc_df = schema_df.ww.loc[:, reversed_cols]
    assert all(reversed_cols == loc_df.columns)
    assert all(reversed_cols == loc_df.ww.types.index)

    getitem_df = schema_df.ww[reversed_cols]
    assert all(reversed_cols == getitem_df.columns)
    assert all(reversed_cols == getitem_df.ww.types.index)


def test_maintain_column_order_disordered_schema(sample_df):
    sample_df.ww.init()
    column_order = list(sample_df.columns)

    scramble_df = sample_df.ww.copy()
    id_col = scramble_df.ww.columns.pop('id')
    scramble_df.ww.columns['id'] = id_col
    assert list(scramble_df.ww.columns.keys()) != column_order

    assert scramble_df.ww.schema == sample_df.ww.schema
    assert all(scramble_df.columns == column_order)
    assert all(scramble_df.ww.types.index == column_order)

    sample_df.ww.init(schema=scramble_df.ww.schema)
    assert all(sample_df.columns == column_order)
    assert all(sample_df.ww.types.index == column_order)


def test_accessor_types(sample_df, sample_inferred_logical_types):
    sample_df.ww.init()

    returned_types = sample_df.ww.types
    assert isinstance(returned_types, pd.DataFrame)
    assert all(returned_types.columns == ['Physical Type', 'Logical Type', 'Semantic Tag(s)'])
    assert returned_types.shape[1] == 3
    assert len(returned_types.index) == len(sample_df.columns)

    correct_physical_types = {name: ltype.primary_dtype for name, ltype in sample_inferred_logical_types.items()}
    if _is_koalas_dataframe(sample_df):
        correct_physical_types['categorical'] = 'string'
    correct_physical_types = pd.Series(list(correct_physical_types.values()),
                                       index=list(correct_physical_types.keys()))

    assert correct_physical_types.equals(returned_types['Physical Type'])
    correct_logical_types = pd.Series([ltype() for ltype in sample_inferred_logical_types.values()],
                                      index=list(sample_inferred_logical_types.keys()))
    assert correct_logical_types.equals(returned_types['Logical Type'])
    correct_semantic_tags = {
        'id': "['numeric']",
        'full_name': "[]",
        'email': "[]",
        'phone_number': "[]",
        'age': "['numeric']",
        'signup_date': "[]",
        'is_registered': "[]",
        'double': "['numeric']",
        'double_with_nan': "['numeric']",
        'integer': "['numeric']",
        'nullable_integer': "['numeric']",
        'boolean': "[]",
        'categorical': "['category']",
        'datetime_with_NaT': "[]",
    }
    correct_semantic_tags = pd.Series(list(correct_semantic_tags.values()),
                                      index=list(correct_semantic_tags.keys()))
    assert correct_semantic_tags.equals(returned_types['Semantic Tag(s)'])


def test_accessor_repr(small_df):
    error = 'Woodwork not initialized for this DataFrame. Initialize by calling DataFrame.ww.init'
    with pytest.raises(WoodworkNotInitError, match=error):
        repr(small_df.ww)
    small_df.ww.init()

    table_repr = repr(small_df.ww)
    expected_repr = '                         Physical Type Logical Type Semantic Tag(s)\nColumn                                                             \nsample_datetime_series  datetime64[ns]     Datetime              []'
    assert table_repr == expected_repr

    table_html_repr = small_df.ww._repr_html_()
    expected_repr = '<table border="1" class="dataframe">\n  <thead>\n    <tr style="text-align: right;">\n      <th></th>\n      <th>Physical Type</th>\n      <th>Logical Type</th>\n      <th>Semantic Tag(s)</th>\n    </tr>\n    <tr>\n      <th>Column</th>\n      <th></th>\n      <th></th>\n      <th></th>\n    </tr>\n  </thead>\n  <tbody>\n    <tr>\n      <th>sample_datetime_series</th>\n      <td>datetime64[ns]</td>\n      <td>Datetime</td>\n      <td>[]</td>\n    </tr>\n  </tbody>\n</table>'
    assert table_html_repr == expected_repr


def test_accessor_repr_empty(empty_df):
    empty_df.ww.init()

    assert repr(empty_df.ww) == 'Empty DataFrame\nColumns: [Physical Type, Logical Type, Semantic Tag(s)]\nIndex: []'
    assert empty_df.ww._repr_html_() == '<table border="1" class="dataframe">\n  <thead>\n    <tr style="text-align: right;">\n      <th></th>\n      <th>Physical Type</th>\n      <th>Logical Type</th>\n      <th>Semantic Tag(s)</th>\n    </tr>\n    <tr>\n      <th>Column</th>\n      <th></th>\n      <th></th>\n      <th></th>\n    </tr>\n  </thead>\n  <tbody>\n  </tbody>\n</table>'


def test_numeric_column_names(sample_df):
    original_df = sample_df.drop('categorical', axis=1)
    numeric_columns = {col_name: i for col_name, i in zip(original_df.columns, range(0, len(original_df.columns)))}

    original_df.ww.init()
    numeric_via_woodwork = original_df.ww.rename(numeric_columns)

    assert numeric_via_woodwork.ww[0].ww._schema == original_df.ww['id'].ww._schema

    numeric_cols_df = original_df.rename(columns=numeric_columns)
    numeric_cols_df.ww.init()

    assert numeric_cols_df.ww == numeric_via_woodwork.ww

    numeric_cols_df.ww.set_index(0)
    assert numeric_cols_df.ww.index == 0
    assert numeric_cols_df.ww.semantic_tags[0] == {'index'}


@patch("woodwork.table_accessor._validate_accessor_params")
def test_validation_methods_called(mock_validate_accessor_params, sample_df):
    assert not mock_validate_accessor_params.called

    not_validated_df = sample_df.copy()
    not_validated_df.ww.init(validate=False, index='id', logical_types={'age': 'Double'})

    assert not mock_validate_accessor_params.called

    validated_df = sample_df.copy()
    validated_df.ww.init(validate=True, index='id', logical_types={'age': 'Double'})

    assert mock_validate_accessor_params.called

    assert validated_df.ww == not_validated_df.ww
    pd.testing.assert_frame_equal(to_pandas(validated_df), to_pandas(not_validated_df))


def test_maintains_set_logical_type(sample_df):
    if _is_koalas_dataframe(sample_df):
        pytest.xfail("Koalas changed dtype on fillna which invalidates schema")
    sample_df.ww.init(logical_types={'age': 'IntegerNullable'})
    assert isinstance(sample_df.ww.logical_types['age'], IntegerNullable)
    new_df = sample_df.ww.fillna({'age': -1})
    assert isinstance(new_df.ww.logical_types['age'], IntegerNullable)


def test_ltype_conversions_nullable_types():
    df = pd.DataFrame({
        'bool': pd.Series([True, False, True], dtype='bool'),
        'bool_null': pd.Series([True, False, pd.NA], dtype='boolean'),
        'int': pd.Series([1, 7, 3], dtype='int64'),
        'int_null': pd.Series([1, 7, pd.NA], dtype='Int64')
    })

    df.ww.init()
    assert isinstance(df.ww.logical_types['bool'], Boolean)
    assert isinstance(df.ww.logical_types['bool_null'], BooleanNullable)
    assert isinstance(df.ww.logical_types['int'], Integer)
    assert isinstance(df.ww.logical_types['int_null'], IntegerNullable)

    # Test valid conversions
    df.ww.set_types({'bool': 'BooleanNullable', 'int': 'IntegerNullable'})
    assert isinstance(df.ww.logical_types['bool'], BooleanNullable)
    assert isinstance(df.ww.logical_types['int'], IntegerNullable)
    df.ww.set_types({'bool': 'Boolean', 'int': 'Integer'})
    assert isinstance(df.ww.logical_types['bool'], Boolean)
    assert isinstance(df.ww.logical_types['int'], Integer)

    # Test invalid conversions
    error_msg = "Error converting datatype for bool_null from type boolean to type bool. " \
        "Please confirm the underlying data is consistent with logical type Boolean."
    with pytest.raises(TypeConversionError, match=error_msg):
        df.ww.set_types({'bool_null': 'Boolean'})

    error_msg = "Error converting datatype for int_null from type Int64 to type int64. " \
        "Please confirm the underlying data is consistent with logical type Integer."
    with pytest.raises(TypeConversionError, match=error_msg):
        df.ww.set_types({'int_null': 'Integer'})



import copy
import warnings
import weakref

import pandas as pd

import woodwork.serialize as serialize
from woodwork.accessor_utils import (
    _check_table_schema,
    _is_dask_dataframe,
    _is_dataframe,
    _is_koalas_dataframe,
    get_invalid_schema_message,
    init_series
)
from woodwork.exceptions import (
    ColumnNotPresentError,
    IndexTagRemovedWarning,
    ParametersIgnoredWarning,
    TypingInfoMismatchWarning
)
from woodwork.indexers import _iLocIndexer, _locIndexer
from woodwork.logical_types import Datetime
from woodwork.statistics_utils import (
    _get_describe_dict,
    _get_mutual_information_dict,
    _get_value_counts
)
from woodwork.table_schema import TableSchema
from woodwork.type_sys.utils import _is_numeric_series, col_is_datetime
from woodwork.utils import (
    _get_column_logical_type,
    _parse_logical_type,
    import_or_none,
    import_or_raise
)

dd = import_or_none('dask.dataframe')
ks = import_or_none('databricks.koalas')


class WoodworkTableAccessor:
    def __init__(self, dataframe):
        self._dataframe_weakref = weakref.ref(dataframe)
        self._schema = None

    def init(self,
             index=None,
             time_index=None,
             logical_types=None,
             already_sorted=False,
             schema=None,
             validate=True,
             use_standard_tags=True,
             **kwargs):
        """Initializes Woodwork typing information for a DataFrame.

        Args:
            index (str, optional): Name of the index column.
            time_index (str, optional): Name of the time index column.
            logical_types (dict[str -> LogicalType]): Dictionary mapping column names in
                the DataFrame to the LogicalType for the column.
            already_sorted (bool, optional): Indicates whether the input DataFrame is already sorted on the time
                index. If False, will sort the dataframe first on the time_index and then on the index (pandas DataFrame
                only). Defaults to False.
            name (str, optional): Name used to identify the DataFrame.
            semantic_tags (dict, optional): Dictionary mapping column names in Woodwork to the
                semantic tags for the column. The keys in the dictionary should be strings
                that correspond to column names. There are two options for
                specifying the dictionary values:
                (str): If only one semantic tag is being set, a single string can be used as a value.
                (list[str] or set[str]): If multiple tags are being set, a list or set of strings can be
                used as the value.
                Semantic tags will be set to an empty set for any column not included in the
                dictionary.
            table_metadata (dict[str -> json serializable], optional): Dictionary containing extra metadata for Woodwork.
            column_metadata (dict[str -> dict[str -> json serializable]], optional): Dictionary mapping column names
                to that column's metadata dictionary.
            use_standard_tags (bool, dict[str -> bool], optional): Determines whether standard semantic tags will be
                added to columns based on the specified logical type for the column.
                If a single boolean is supplied, will apply the same use_standard_tags value to all columns.
                A dictionary can be used to specify ``use_standard_tags`` values for individual columns.
                Unspecified columns will use the default value. Defaults to True.
            column_descriptions (dict[str -> str], optional): Dictionary mapping column names to column descriptions.
            column_origins (str, dict[str -> str], optional): Origin of each column. If a string is supplied, it is
                used as the origin for all columns. A dictionary can be used to set origins for individual columns.
            schema (Woodwork.TableSchema, optional): Typing information to use for the DataFrame instead of performing inference.
                Any other arguments provided will be ignored. Note that any changes made to the schema object after
                initialization will propagate to the DataFrame. Similarly, to avoid unintended typing information changes,
                the same schema object should not be shared between DataFrames.
            validate (bool, optional): Whether parameter and data validation should occur. Defaults to True. Warning:
                Should be set to False only when parameters and data are known to be valid.
                Any errors resulting from skipping validation with invalid inputs may not be easily understood.
        """
        if validate:
            _validate_accessor_params(self._dataframe, index, time_index, logical_types, schema, use_standard_tags)
        if schema is not None:
            self._schema = schema
            extra_params = []
            if index is not None:
                extra_params.append('index')
            if time_index is not None:
                extra_params.append('time_index')
            if logical_types is not None:
                extra_params.append('logical_types')
            if already_sorted:
                extra_params.append('already_sorted')
            if not use_standard_tags or isinstance(use_standard_tags, dict):
                extra_params.append('use_standard_tags')
            for key in kwargs:
                extra_params.append(key)
            if extra_params:
                warnings.warn("A schema was provided and the following parameters were ignored: " + ", ".join(extra_params), ParametersIgnoredWarning)

        else:
            # Perform type inference and update underlying data
            parsed_logical_types = {}
            for name in self._dataframe.columns:
                series = self._dataframe[name]

                logical_type = None
                if logical_types:
                    logical_type = logical_types.get(name)

                logical_type = _get_column_logical_type(series, logical_type, name)
                parsed_logical_types[name] = logical_type

                updated_series = logical_type.transform(series)
                if updated_series is not series:
                    self._dataframe[name] = updated_series

            column_names = list(self._dataframe.columns)

            # TableSchema uses a different default for use_standard_tags so we need to define it here
            if isinstance(use_standard_tags, bool):
                use_standard_tags = {col_name: use_standard_tags for col_name in column_names}
            else:
                use_standard_tags = {**{col_name: True for col_name in column_names}, **use_standard_tags}

            self._schema = TableSchema(column_names=column_names,
                                       logical_types=parsed_logical_types,
                                       index=index,
                                       time_index=time_index,
                                       validate=validate,
                                       use_standard_tags=use_standard_tags,
                                       **kwargs)

            self._set_underlying_index()
            if self._schema.time_index is not None:
                self._sort_columns(already_sorted)

    def __eq__(self, other, deep=True):
        if not self._schema.__eq__(other.ww._schema, deep=deep):
            return False

        # Only check pandas DataFrames for equality
        if deep and isinstance(self._dataframe, pd.DataFrame) and isinstance(other.ww._dataframe, pd.DataFrame):
            return self._dataframe.equals(other.ww._dataframe)
        return True

    @_check_table_schema
    def __getattr__(self, attr):
        # Called if method is not present on the Accessor
        # If the method is present on TableSchema, uses that method.
        # If the method is present on DataFrame, uses that method.
        if hasattr(self._schema, attr):
            return self._make_schema_call(attr)
        if hasattr(self._dataframe, attr):
            return self._make_dataframe_call(attr)
        else:
            raise AttributeError(f"Woodwork has no attribute '{attr}'")

    @_check_table_schema
    def __getitem__(self, key):
        if isinstance(key, list):
            columns = set(self._dataframe.columns)
            diff = list(set(key).difference(columns))

            if diff:
                raise ColumnNotPresentError(sorted(diff))

            return self._get_subset_df_with_schema(key)

        if key not in self._dataframe:
            raise ColumnNotPresentError(key)

        series = self._dataframe[key]
        column = copy.deepcopy(self._schema.columns[key])

        series.ww.init(schema=column, validate=False)

        return series

    def __setitem__(self, col_name, column):
        series = tuple(pkg.Series for pkg in (pd, dd, ks) if pkg)
        if not isinstance(column, series):
            raise ValueError('New column must be of Series type')

        if column.ww.schema is not None and 'index' in column.ww.semantic_tags:
            warnings.warn(f'Cannot add "index" tag on {col_name} directly to the DataFrame. The "index" tag has been removed from {col_name}. To set this column as a Woodwork index, please use df.ww.set_index', IndexTagRemovedWarning)
            column.ww.set_semantic_tags(column.ww.semantic_tags - {'index'})

        # Don't allow reassigning of index or time index with setitem
        if self.index == col_name:
            raise KeyError('Cannot reassign index. Change column name and then use df.ww.set_index to reassign index.')
        if self.time_index == col_name:
            raise KeyError('Cannot reassign time index. Change column name and then use df.ww.set_time_index to reassign time index.')

        if column.ww._schema is None:
            column = init_series(column, use_standard_tags=True)

        self._dataframe[col_name] = column
        self._schema.columns[col_name] = column.ww._schema

    def __repr__(self):
        """A string representation of a Woodwork table containing typing information"""
        return repr(self._get_typing_info())

    def _repr_html_(self):
        """An HTML representation of a Woodwork table for IPython.display in Jupyter Notebooks
        containing typing information and a preview of the data."""
        return self._get_typing_info().to_html()

    @_check_table_schema
    def _get_typing_info(self):
        """Creates a DataFrame that contains the typing information for a Woodwork table."""
        typing_info = self._schema._get_typing_info().copy()
        typing_info.insert(0, 'Physical Type', pd.Series(self.physical_types))
        # Maintain the same column order used in the DataFrame
        typing_info = typing_info.loc[list(self._dataframe.columns), :]
        return typing_info

    @property
    @_check_table_schema
    def name(self):
        """Name of the DataFrame"""
        return self._schema.name

    @name.setter
    @_check_table_schema
    def name(self, name):
        """Set name of the DataFrame"""
        self._schema.name = name

    @property
    @_check_table_schema
    def metadata(self):
        """Metadata of the DataFrame"""
        return self._schema.metadata

    @metadata.setter
    @_check_table_schema
    def metadata(self, metadata):
        """Set metadata of the DataFrame"""
        self._schema.metadata = metadata

    @property
    def _dataframe(self):
        return self._dataframe_weakref()

    @property
    @_check_table_schema
    def iloc(self):
        """
        Integer-location based indexing for selection by position.
        ``.iloc[]`` is primarily integer position based (from ``0`` to
        ``length-1`` of the axis), but may also be used with a boolean array.

        If the selection result is a DataFrame or Series, Woodwork typing
        information will be initialized for the returned object when possible.

        Allowed inputs are:
            An integer, e.g. ``5``.
            A list or array of integers, e.g. ``[4, 3, 0]``.
            A slice object with ints, e.g. ``1:7``.
            A boolean array.
            A ``callable`` function with one argument (the calling Series, DataFrame
            or Panel) and that returns valid output for indexing (one of the above).
            This is useful in method chains, when you don't have a reference to the
            calling object, but would like to base your selection on some value.
        """
        return _iLocIndexer(self._dataframe)

    @property
    @_check_table_schema
    def loc(self):
        """
        Access a group of rows by label(s) or a boolean array.

        ``.loc[]`` is primarily label based, but may also be used with a
        boolean array.

        If the selection result is a DataFrame or Series, Woodwork typing
        information will be initialized for the returned object when possible.

        Allowed inputs are:
            A single label, e.g. ``5`` or ``'a'``, (note that ``5`` is
            interpreted as a *label* of the index, and **never** as an
            integer position along the index).
            A list or array of labels, e.g. ``['a', 'b', 'c']``.
            A slice object with labels, e.g. ``'a':'f'``.
            A boolean array of the same length as the axis being sliced,
            e.g. ``[True, False, True]``.
            An alignable boolean Series. The index of the key will be aligned before
            masking.
            An alignable Index. The Index of the returned selection will be the input.
            A ``callable`` function with one argument (the calling Series or
            DataFrame) and that returns valid output for indexing (one of the above)
        """
        return _locIndexer(self._dataframe)

    @property
    def schema(self):
        """A copy of the Woodwork typing information for the DataFrame."""
        if self._schema:
            return copy.deepcopy(self._schema)

    @property
    @_check_table_schema
    def physical_types(self):
        """A dictionary containing physical types for each column"""
        return {col_name: self._schema.logical_types[col_name]._get_valid_dtype(type(self._dataframe[col_name])) for col_name in self._dataframe.columns}

    @property
    @_check_table_schema
    def types(self):
        """DataFrame containing the physical dtypes, logical types and semantic
        tags for the schema."""
        return self._get_typing_info()

    @property
    @_check_table_schema
    def logical_types(self):
        """A dictionary containing logical types for each column"""
        return self._schema.logical_types

    @property
    @_check_table_schema
    def semantic_tags(self):
        """A dictionary containing semantic tags for each column"""
        return self._schema.semantic_tags

    @property
    @_check_table_schema
    def index(self):
        """The index column for the table"""
        return self._schema.index

    @property
    @_check_table_schema
    def time_index(self):
        """The time index column for the table"""
        return self._schema.time_index

    @property
    @_check_table_schema
    def use_standard_tags(self):
        """A dictionary containing the use_standard_tags setting for each column in the table"""
        return self._schema.use_standard_tags

    @_check_table_schema
    def set_index(self, new_index):
        """Sets the index column of the DataFrame. Adds the 'index' semantic tag to the column
        and clears the tag from any previously set index column.

        Setting a column as the index column will also cause any previously set standard
        tags for the column to be removed.

        Clears the DataFrame's index by passing in None.

        Args:
            new_index (str): The name of the column to set as the index
        """
        self._schema.set_index(new_index)

        if self._schema.index is not None:
            _check_index(self._dataframe, self._schema.index)
        self._set_underlying_index()

    @_check_table_schema
    def set_time_index(self, new_time_index):
        """Set the time index. Adds the 'time_index' semantic tag to the column and
        clears the tag from any previously set index column

        Args:
            new_time_index (str): The name of the column to set as the time index.
                If None, will remove the time_index.
        """
        self._schema.set_time_index(new_time_index)

    @_check_table_schema
    def set_types(self, logical_types=None, semantic_tags=None, retain_index_tags=True):
        """Update the logical type and semantic tags for any columns names in the provided types dictionaries,
        updating the Woodwork typing information for the DataFrame.

        Args:
            logical_types (dict[str -> str], optional): A dictionary defining the new logical types for the
                specified columns.
            semantic_tags (dict[str -> str/list/set], optional): A dictionary defining the new semantic_tags for the
                specified columns.
            retain_index_tags (bool, optional): If True, will retain any index or time_index
                semantic tags set on the column. If False, will replace all semantic tags any time a column's
                semantic tags or logical type changes. Defaults to True.
        """
        logical_types = logical_types or {}
        logical_types = {col_name: _parse_logical_type(ltype, col_name) for col_name, ltype in logical_types.items()}

        self._schema.set_types(logical_types=logical_types,
                               semantic_tags=semantic_tags,
                               retain_index_tags=retain_index_tags)
        # go through changed ltypes and update dtype if necessary
        for col_name, logical_type in logical_types.items():
            series = self._dataframe[col_name]
            updated_series = logical_type.transform(series)
            if updated_series is not series:
                self._dataframe[col_name] = updated_series

    @_check_table_schema
    def select(self, include=None, exclude=None, return_schema=False):
        """Create a DataFrame with Woodwork typing information initialized
        that includes only columns whose Logical Type and semantic tags match
        conditions specified in the list of types and tags to include or exclude.
        Values for both ``include`` and ``exclude`` cannot be provided in a
        single call.

        If no matching columns are found, an empty DataFrame will be returned.

        Args:
            include (str or LogicalType or list[str or LogicalType]): Logical
                types, semantic tags to include in the DataFrame.
            exclude (str or LogicalType or list[str or LogicalType]): Logical
                types, semantic tags to exclude from the DataFrame.
            return_schema (boolen): If True, return only the schema for the
                matching columns. Defaults to False

        Returns:
            DataFrame: The subset of the original DataFrame that matches the
            conditions specified by ``include`` or ``exclude``. Has Woodwork
            typing information initialized.
        """
        if include is not None and exclude is not None:
            raise ValueError("Cannot specify values for both 'include' and 'exclude' in a single call.")
        if include is None and exclude is None:
            raise ValueError("Must specify values for either 'include' or 'exclude'.")

        cols_to_include = self._schema._filter_cols(include, exclude)

        if return_schema:
            return self._schema._get_subset_schema(cols_to_include)
        return self._get_subset_df_with_schema(cols_to_include)

    @_check_table_schema
    def add_semantic_tags(self, semantic_tags):
        """Adds specified semantic tags to columns, updating the Woodwork typing information.
        Will retain any previously set values.

        Args:
            semantic_tags (dict[str -> str/list/set]): A dictionary mapping the columns
                in the DataFrame to the tags that should be added to the column's semantic tags
        """
        self._schema.add_semantic_tags(semantic_tags)

    @_check_table_schema
    def remove_semantic_tags(self, semantic_tags):
        """Remove the semantic tags for any column names in the provided semantic_tags
        dictionary, updating the Woodwork typing information. Including `index` or `time_index`
        tags will set the Woodwork index or time index to None for the DataFrame.

        Args:
            semantic_tags (dict[str -> str/list/set]): A dictionary mapping the columns
                in the DataFrame to the tags that should be removed from the column's semantic tags
        """
        self._schema.remove_semantic_tags(semantic_tags)

    @_check_table_schema
    def reset_semantic_tags(self, columns=None, retain_index_tags=False):
        """Reset the semantic tags for the specified columns to the default values.
        The default values will be either an empty set or a set of the standard tags
        based on the column logical type, controlled by the use_standard_tags property on each column.
        Column names can be provided as a single string, a list of strings or a set of strings.
        If columns is not specified, tags will be reset for all columns.

        Args:
            columns (str/list/set, optional): The columns for which the semantic tags should be reset.
            retain_index_tags (bool, optional): If True, will retain any index or time_index
                semantic tags set on the column. If False, will clear all semantic tags. Defaults to
                False.
        """
        self._schema.reset_semantic_tags(columns=columns, retain_index_tags=retain_index_tags)

    @_check_table_schema
    def to_dictionary(self):
        """Get a dictionary representation of the Woodwork typing information.

        Returns:
            dict: Description of the typing information.
        """
        return serialize.typing_info_to_dict(self._dataframe)

    @_check_table_schema
    def to_disk(self, path, format='csv', compression=None, profile_name=None, **kwargs):
        """Write Woodwork table to disk in the format specified by `format`, location specified by `path`.
            Path could be a local path or an S3 path.
            If writing to S3 a tar archive of files will be written.

            Note:
                As the engine `fastparquet` cannot handle nullable pandas dtypes, `pyarrow` will be used
                for serialization to parquet.


            Args:
                path (str) : Location on disk to write to (will be created as a directory)
                format (str) : Format to use for writing Woodwork data. Defaults to csv. Possible values are: {'csv', 'pickle', 'parquet'}.
                compression (str) : Name of the compression to use. Possible values are: {'gzip', 'bz2', 'zip', 'xz', None}.
                profile_name (str) : Name of AWS profile to use, False to use an anonymous profile, or None.
                kwargs (keywords) : Additional keyword arguments to pass as keywords arguments to the underlying serialization method or to specify AWS profile.
        """
        if format == 'csv':
            default_csv_kwargs = {'sep': ',', 'encoding': 'utf-8', 'engine': 'python', 'index': False}
            if _is_koalas_dataframe(self._dataframe):
                default_csv_kwargs['multiline'] = True
                default_csv_kwargs['ignoreLeadingWhitespace'] = False
                default_csv_kwargs['ignoreTrailingWhitespace'] = False
            kwargs = {**default_csv_kwargs, **kwargs}
        elif format in ['parquet', 'orc']:
            import_error_message = (
                f"The pyarrow library is required to serialize to {format}.\n"
                "Install via pip:\n"
                "    pip install pyarrow\n"
                "Install via conda:\n"
                "   conda install pyarrow -c conda-forge"
            )
            import_or_raise('pyarrow', import_error_message)
            kwargs['engine'] = 'pyarrow'
        serialize.write_woodwork_table(self._dataframe, path, format=format,
                                       compression=compression, profile_name=profile_name,
                                       **kwargs)

    def _sort_columns(self, already_sorted):
        if _is_dask_dataframe(self._dataframe) or _is_koalas_dataframe(self._dataframe):
            already_sorted = True  # Skip sorting for Dask and Koalas input
        if not already_sorted:
            sort_cols = [self._schema.time_index, self._schema.index]
            if self._schema.index is None:
                sort_cols = [self._schema.time_index]
            self._dataframe.sort_values(sort_cols, inplace=True)

    def _set_underlying_index(self):
        """Sets the index of the underlying DataFrame to match the index column as
        specified by the TableSchema. Does not change the underlying index if no Woodwork index is
        specified. Only sets underlying index for pandas DataFrames.
        """
        if isinstance(self._dataframe, pd.DataFrame) and self._schema.index is not None:
            self._dataframe.set_index(self._schema.index, drop=False, inplace=True)
            # Drop index name to not overlap with the original column
            self._dataframe.index.name = None

    def _make_schema_call(self, attr):
        """Forwards the requested attribute onto the schema object.
        Results are that of the Woodwork.TableSchema class."""
        schema_attr = getattr(self._schema, attr)

        if callable(schema_attr):
            def wrapper(*args, **kwargs):
                return schema_attr(*args, **kwargs)
            return wrapper
        return schema_attr

    def _make_dataframe_call(self, attr):
        """Forwards the requested attribute onto the dataframe object. Intercepts return value,
        attempting to initialize Woodwork with the current schema when a new DataFrame is returned.
        Confirms schema is still valid for the original DataFrame."""
        dataframe_attr = getattr(self._dataframe, attr)

        if callable(dataframe_attr):
            def wrapper(*args, **kwargs):
                # Make DataFrame call and intercept the result
                result = dataframe_attr(*args, **kwargs)

                # Try to initialize Woodwork with the existing schema
                if _is_dataframe(result):
                    invalid_schema_message = get_invalid_schema_message(result, self._schema)
                    if invalid_schema_message:
                        warnings.warn(TypingInfoMismatchWarning().get_warning_message(attr, invalid_schema_message, 'DataFrame'),
                                      TypingInfoMismatchWarning)
                    else:
                        copied_schema = self.schema
                        result.ww.init(schema=copied_schema, validate=False)
                else:
                    # Confirm that the schema is still valid on original DataFrame
                    # Important for inplace operations
                    invalid_schema_message = get_invalid_schema_message(self._dataframe, self._schema)

                    if invalid_schema_message:
                        warnings.warn(TypingInfoMismatchWarning().get_warning_message(attr, invalid_schema_message, 'DataFrame'),
                                      TypingInfoMismatchWarning)
                        self._schema = None

                # Always return the results of the DataFrame operation whether or not Woodwork is initialized
                return result
            return wrapper
        # Directly return non-callable DataFrame attributes
        return dataframe_attr

    def _get_subset_df_with_schema(self, cols_to_include, inplace=False):
        """Creates a new DataFrame from a list of column names with Woodwork initialized,
        retaining all typing information and maintaining the DataFrame's column order."""
        if inplace:
            if _is_dask_dataframe(self._dataframe):
                raise ValueError('Drop inplace not supported for Dask')
            if _is_koalas_dataframe(self._dataframe):
                raise ValueError('Drop inplace not supported for Koalas')

        assert all([col_name in self._schema.columns for col_name in cols_to_include])

        new_schema = self._schema._get_subset_schema(cols_to_include)
        if inplace:
            cols_to_drop = [col_name for col_name in self._dataframe.columns if col_name not in cols_to_include]
            self._dataframe.drop(cols_to_drop, axis='columns', inplace=True)
            self.init(schema=new_schema, validate=False)
            return
        new_df = self._dataframe[cols_to_include]
        new_df.ww.init(schema=new_schema, validate=False)

        return new_df

    @_check_table_schema
    def pop(self, column_name):
        """Return a Series with Woodwork typing information and remove it from the DataFrame.

        Args:
            column (str): Name of the column to pop.

        Returns:
            Series: Popped series with Woodwork initialized
        """
        if column_name not in self._dataframe.columns:
            raise ColumnNotPresentError(column_name)

        series = self._dataframe.pop(column_name)

        # Initialize Woodwork typing info for series
        series.ww.init(schema=self.schema.columns[column_name], validate=False)

        # Update schema to not include popped column
        del self._schema.columns[column_name]

        return series

    @_check_table_schema
    def drop(self, columns, inplace=False):
        """Drop specified columns from a DataFrame.

        Args:
            columns (str or list[str]): Column name or names to drop. Must be present in the DataFrame.
            inplace (bool): If False, return a copy. Otherwise, do operation inplace and return None.

        Returns:
            DataFrame or None: DataFrame with the specified columns removed, maintaining Woodwork typing information or None if inplace=True. Only possible for pandas dataframes.

        Note:
            This method is used for removing columns only. To remove rows with ``drop``, go through the
            DataFrame directly and then reinitialize Woodwork with ``DataFrame.ww.init``
            instead of calling ``DataFrame.ww.drop``.
        """
        if not isinstance(columns, (list, set)):
            columns = [columns]

        not_present = [col for col in columns if col not in self._dataframe.columns]
        if not_present:
            raise ColumnNotPresentError(not_present)
        return self._get_subset_df_with_schema([col for col in self._dataframe.columns if col not in columns], inplace=inplace)

    @_check_table_schema
    def rename(self, columns, inplace=False):
        """Renames columns in a DataFrame, maintaining Woodwork typing information.

        Args:
            columns (dict[str -> str]): A dictionary mapping current column names to new column names.
            inplace (bool): If False, return a copy. Otherwise, do operation inplace and return None.

        Returns:
            DataFrame or None: DataFrame with the specified columns renamed, maintaining Woodwork typing information or None if inplace=True. Only possible for pandas dataframes.
        """

        new_schema = self._schema.rename(columns)
        if inplace:
            if _is_dask_dataframe(self._dataframe):
                raise ValueError('Rename inplace not supported for Dask')
            if _is_koalas_dataframe(self._dataframe):
                raise ValueError('Rename inplace not supported for Koalas')
            self._dataframe.rename(columns=columns, inplace=True)
            self.init(schema=new_schema)
            return
        new_df = self._dataframe.rename(columns=columns)

        new_df.ww.init(schema=new_schema)
        return new_df

    @_check_table_schema
    def mutual_information_dict(self, num_bins=10, nrows=None, include_index=False, callback=None):
        """
        Calculates mutual information between all pairs of columns in the DataFrame that
        support mutual information. Logical Types that support mutual information are as
        follows:  Age, AgeNullable, Boolean, BooleanNullable, Categorical, CountryCode, Datetime, Double,
        Integer, IntegerNullable, Ordinal, PostalCode, and SubRegionCode

        Args:
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical.
            nrows (int): The number of rows to sample for when determining mutual info.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for mutual information calculations.
                If False, the index column will not have mutual information calculated for it.
                Defaults to False.
            callback (callable, optional): function to be called with incremental updates. Has the following parameters:

                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call

        Returns:
            list(dict): A list containing dictionaries that have keys `column_1`,
            `column_2`, and `mutual_info` that is sorted in decending order by mutual info.
            Mutual information values are between 0 (no mutual information) and 1
            (perfect dependency).
        """
        return _get_mutual_information_dict(self._dataframe, num_bins, nrows, include_index, callback)

    def mutual_information(self, num_bins=10, nrows=None, include_index=False, callback=None):
        """Calculates mutual information between all pairs of columns in the DataFrame that
        support mutual information. Logical Types that support mutual information are as
        follows:  Age, AgeNullable, Boolean, BooleanNullable, Categorical, CountryCode, Datetime, Double,
        Integer, IntegerNullable, Ordinal, PostalCode, and SubRegionCode

        Args:
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical.
            nrows (int): The number of rows to sample for when determining mutual info.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for mutual information calculations.
                If False, the index column will not have mutual information calculated for it.
                Defaults to False.
            callback (callable, optional): function to be called with incremental updates. Has the following parameters:

                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call

        Returns:
            pd.DataFrame: A DataFrame containing mutual information with columns `column_1`,
            `column_2`, and `mutual_info` that is sorted in decending order by mutual info.
            Mutual information values are between 0 (no mutual information) and 1
            (perfect dependency).
        """
        mutual_info = self.mutual_information_dict(num_bins, nrows, include_index, callback)
        return pd.DataFrame(mutual_info)

    @_check_table_schema
    def describe_dict(self, include=None, callback=None, extra_stats=False, bins=10, top_x=10, recent_x=10):
        """Calculates statistics for data contained in the DataFrame.

        Args:
            include (list[str or LogicalType], optional): filter for what columns to include in the
                statistics returned. Can be a list of column names, semantic tags, logical types, or a list
                combining any of the three. It follows the most broad specification. Favors logical types
                then semantic tag then column name. If no matching columns are found, an empty DataFrame
                will be returned.
            callback (callable, optional): function to be called with incremental updates. Has the following parameters:

                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call

            extra_stats (bool): If True, will calculate a histogram for numeric columns, top values
                for categorical columns and value counts for the most recent values in datetime columns. Will also
                calculate value counts within the range of values present for integer columns if the range of
                values present is less than or equal to than the number of bins used to compute the histogram.
                Output can be controlled by bins, top_x and recent_x parameters.
            bins (int): Number of bins to use when calculating histogram for numeric columns. Defaults to 10.
                Will be ignored unless extra_stats=True.
            top_x (int): Number of items to return when getting the most frequently occurring values for categorical
                columns. Defaults to 10. Will be ignored unless extra_stats=True.
            recent_x (int): Number of values to return when calculating value counts for the most recent dates in
                datetime columns. Defaults to 10. Will be ignored unless extra_stats=True.

        Returns:
            dict[str -> dict]: A dictionary with a key for each column in the data or for each column
            matching the logical types, semantic tags or column names specified in ``include``, paired
            with a value containing a dictionary containing relevant statistics for that column.
        """
        return _get_describe_dict(self._dataframe, include=include, callback=callback,
                                  extra_stats=extra_stats, bins=bins, top_x=top_x, recent_x=recent_x)

    def describe(self, include=None, callback=None):
        """Calculates statistics for data contained in the DataFrame.

        Args:
            include (list[str or LogicalType], optional): filter for what columns to include in the
                statistics returned. Can be a list of column names, semantic tags, logical types, or a list
                combining any of the three. It follows the most broad specification. Favors logical types
                then semantic tag then column name. If no matching columns are found, an empty DataFrame
                will be returned.
            callback (callable, optional): function to be called with incremental updates. Has the following parameters:

                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call

        Returns:
            pd.DataFrame: A Dataframe containing statistics for the data or the subset of the original
            DataFrame that contains the logical types, semantic tags, or column names specified
            in ``include``.
        """
        results = self.describe_dict(include=include, callback=callback)
        index_order = [
            'physical_type',
            'logical_type',
            'semantic_tags',
            'count',
            'nunique',
            'nan_count',
            'mean',
            'mode',
            'std',
            'min',
            'first_quartile',
            'second_quartile',
            'third_quartile',
            'max',
            'num_true',
            'num_false',
        ]
        return pd.DataFrame(results).reindex(index_order)

    @_check_table_schema
    def value_counts(self, ascending=False, top_n=10, dropna=False):
        """Returns a list of dictionaries with counts for the most frequent values in each column (only
            for columns with `category` as a standard tag).


        Args:
            ascending (bool): Defines whether each list of values should be sorted most frequent
                to least frequent value (False), or least frequent to most frequent value (True).
                Defaults to False.

            top_n (int): the number of top values to retrieve. Defaults to 10.

            dropna (bool): determines whether to remove NaN values when finding frequency. Defaults
                to False.

        Returns:
            list(dict): a list of dictionaries for each categorical column with keys `count`
            and `value`.
        """
        return _get_value_counts(self._dataframe, ascending, top_n, dropna)


def _validate_accessor_params(dataframe, index, time_index, logical_types, schema, use_standard_tags):
    _check_unique_column_names(dataframe)
    _check_use_standard_tags(use_standard_tags)
    if schema is not None:
        _check_schema(dataframe, schema)
    else:
        # We ignore these parameters if a schema is passed
        if index is not None:
            _check_index(dataframe, index)
        if logical_types:
            _check_logical_types(dataframe.columns, logical_types)
        if time_index is not None:
            datetime_format = None
            logical_type = None
            if logical_types is not None and time_index in logical_types:
                logical_type = logical_types[time_index]
                if type(logical_types[time_index]) == Datetime:
                    datetime_format = logical_types[time_index].datetime_format

            _check_time_index(dataframe, time_index, datetime_format=datetime_format, logical_type=logical_type)


def _check_unique_column_names(dataframe):
    if not dataframe.columns.is_unique:
        raise IndexError('Dataframe cannot contain duplicate columns names')


def _check_index(dataframe, index):
    if index not in dataframe.columns:
        # User specifies an index that is not in the dataframe
        raise ColumnNotPresentError(f'Specified index column `{index}` not found in dataframe')
    if index is not None and isinstance(dataframe, pd.DataFrame) and not dataframe[index].is_unique:
        # User specifies an index that is in the dataframe but not unique
        # Does not check for Dask as Dask does not support is_unique
        raise IndexError('Index column must be unique')


def _check_time_index(dataframe, time_index, datetime_format=None, logical_type=None):
    if time_index not in dataframe.columns:
        raise ColumnNotPresentError(f'Specified time index column `{time_index}` not found in dataframe')
    if not (_is_numeric_series(dataframe[time_index], logical_type) or
            col_is_datetime(dataframe[time_index], datetime_format=datetime_format)):
        raise TypeError('Time index column must contain datetime or numeric values')


def _check_logical_types(dataframe_columns, logical_types):
    if not isinstance(logical_types, dict):
        raise TypeError('logical_types must be a dictionary')
    cols_not_found = set(logical_types.keys()).difference(set(dataframe_columns))
    if cols_not_found:
        raise ColumnNotPresentError('logical_types contains columns that are not present in '
                                    f'dataframe: {sorted(list(cols_not_found))}')


def _check_schema(dataframe, schema):
    if not isinstance(schema, TableSchema):
        raise TypeError('Provided schema must be a Woodwork.TableSchema object.')
    invalid_schema_message = get_invalid_schema_message(dataframe, schema)
    if invalid_schema_message:
        raise ValueError(f'Woodwork typing information is not valid for this DataFrame: {invalid_schema_message}')


def _check_use_standard_tags(use_standard_tags):
    if not isinstance(use_standard_tags, (bool, dict)):
        raise TypeError('use_standard_tags must be a dictionary or a boolean')


@pd.api.extensions.register_dataframe_accessor('ww')
class PandasTableAccessor(WoodworkTableAccessor):
    pass


if dd:
    @dd.extensions.register_dataframe_accessor('ww')
    class DaskTableAccessor(WoodworkTableAccessor):
        pass


if ks:
    from databricks.koalas.extensions import register_dataframe_accessor

    @register_dataframe_accessor('ww')
    class KoalasTableAccessor(WoodworkTableAccessor):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if not ks.get_option('compute.ops_on_diff_frames'):
                ks.set_option('compute.ops_on_diff_frames', True)
